/*   FILE: GraphicsManager.java
 *   DATE OF CREATION:   Mon Nov 27 08:30:31 2006
 *   Copyright (c) INRIA, 2006-2011. All Rights Reserved
 *   Licensed under the GNU LGPL. For full terms see the file COPYING.
 *
 *   $Id$
 */

package net.claribole.zgrviewer;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.util.*;

import javax.swing.*;

import fr.inria.zvtm.animation.*;
import fr.inria.zvtm.animation.interpolation.*;
import fr.inria.zvtm.engine.*;
import fr.inria.zvtm.engine.portals.*;
import fr.inria.zvtm.event.*;
import fr.inria.zvtm.glyphs.*;
import fr.inria.zvtm.lens.*;
import fr.inria.zvtm.svg.*;

/* Multiscale navigation manager */

public class GraphicsManager implements ComponentListener, CameraListener, Java2DPainter {

    static final Color FADE_COLOR = Color.WHITE;

    public VirtualSpaceManager vsm;
    AnimationManager animator;
    VirtualSpace mSpace; // virtual space containing graph
    VirtualSpace mnSpace; // virtual space containing pie menu
    VirtualSpace rSpace; // virtual space containing rectangle representing region seen through main camera (used in overview)
    static final String mainSpaceName = "graphSpace";
    static final String menuSpace = "menuSpace";
    /*name of the VTM virtual space holding the rectangle delimiting the region seen by main view in radar view*/
    static final String rdRegionVirtualSpaceName = "radarSpace";
    /*represents the region seen by main view in the radar view*/
    VRectangle observedRegion;

    public View mainView;
    View rView;
    static final String RADAR_VIEW_NAME = "Overview";
    Camera mainCamera;
    JPanel mainViewPanel;

    PeriodicActionManager paMngr;

    GeometryEditor geom;

    /*dimensions of zoomable panel*/
    int panelWidth, panelHeight;

    public int getPanelWidth() {
        return this.panelWidth;
    }

    public int getPanelHeight() {
        return this.panelHeight;
    }

    /* misc. lens settings */
    Lens lens;
    SCBLens fLens;
    public static final int LENS_R1 = 100;
    public static final int LENS_R2 = 50;
    public static final int WHEEL_ANIM_TIME = 50;
    public static final int LENS_ANIM_TIME = 300;
    public static final double DEFAULT_MAG_FACTOR = 4.0;
    static double MAG_FACTOR = DEFAULT_MAG_FACTOR;
    static double INV_MAG_FACTOR = 1 / MAG_FACTOR;
    public static final float WHEEL_MM_STEP = 1.0f;
    public static final float MAX_MAG_FACTOR = 12.0f;

    /* DragMag */
    public static final int DM_PORTAL_WIDTH = 200;
    public static final int DM_PORTAL_HEIGHT = 200;
    public static final int DM_PORTAL_INITIAL_X_OFFSET = 150;
    public static final int DM_PORTAL_INITIAL_Y_OFFSET = 150;
    public static final int DM_PORTAL_ANIM_TIME = 150;
    public static final Color DM_COLOR = Color.RED;
    Camera dmCamera;
    DraggableCameraPortal dmPortal;
    VRectangle magWindow;
    int magWindowW, magWindowN, magWindowE, magWindowS;
    boolean paintLinks = false;

    static final float FLOOR_ALTITUDE = -90.0f;

    /*translation constants*/
    public static final short MOVE_UP = 0;
    public static final short MOVE_DOWN = 1;
    public static final short MOVE_LEFT = 2;
    public static final short MOVE_RIGHT = 3;
    public static final short MOVE_UP_LEFT = 4;
    public static final short MOVE_UP_RIGHT = 5;
    public static final short MOVE_DOWN_LEFT = 6;
    public static final short MOVE_DOWN_RIGHT = 7;

    ToolPalette tp;

    BaseEventHandler meh;
    RadarEvtHdlr reh;

    ConfigManager cfgMngr;

    /*remember previous camera locations so that we can get back*/
    static final int MAX_PREV_LOC = 10;
    Vector previousLocations;

    public static final int NO_LENS = 0;
    public static final int ZOOMIN_LENS = 1;
    public static final int ZOOMOUT_LENS = -1;
    int lensType = NO_LENS;

    /*quick search variables*/
    int searchIndex = 0;
    String lastSearchedString = "";
    Vector matchingList = new Vector();

    /* Some versions of GraphViz generate a rectangle representing the graph's bounding box.
     * We don't want this rectangle neither to be highlighted nor sensitive to zoom clicks,
     * so we try to find it after the SVG has been parsed, and we give it special treatment.
     * null if no bounding box was found in the SVG.
     */
    VRectangleOr boundingBox;

    ZGRGlassPane gp;

    ZGRApplication zapp;

    /* logical structure available only if a recent version of GraphViz
       was used to generate the SVG file (e.g. 2.13, maybe earlier) */
    LogicalStructure lstruct = null;

    GraphicsManager(final ZGRApplication za) {
        this.zapp = za;
        this.geom = new GeometryEditor(this);
    }

    Vector createZVTMelements(final boolean applet) {
        this.vsm = VirtualSpaceManager.INSTANCE;
        VText.setMainFont(ConfigManager.defaultFont);
        Glyph.setDefaultCursorInsideHighlightColor(ConfigManager.HIGHLIGHT_COLOR);
        this.animator = this.vsm.getAnimationManager();
        //vsm.setDebug(true);
        this.mSpace = this.vsm.addVirtualSpace(mainSpaceName);
        // camera #0 for main view
        this.mainCamera = this.mSpace.addCamera();
        this.mainCamera.setZoomFloor(-90);
        // camera #1 for radar view
        this.mSpace.addCamera();
        this.mnSpace = this.vsm.addVirtualSpace(menuSpace);
        // camera for pie menu
        this.mnSpace.addCamera().setAltitude(10);
        this.rSpace = this.vsm.addVirtualSpace(rdRegionVirtualSpaceName);
        // camera for rectangle representing region seen in main viewport (in overview)
        this.rSpace.addCamera();
        // DragMag portal camera (camera #2)
        this.dmCamera = this.mSpace.addCamera();
        SIRectangle seg1;
        SIRectangle seg2;
        this.observedRegion =
            new VRectangle(
                0,
                0,
                0,
                10,
                10,
                ConfigManager.OBSERVED_REGION_COLOR,
                ConfigManager.OBSERVED_REGION_CROSSHAIR_COLOR,
                0.5f);
        //500 should be sufficient as the radar window is
        seg1 = new SIRectangle(0, 0, 0, 0, 500, ConfigManager.OBSERVED_REGION_CROSSHAIR_COLOR);
        //not resizable and is 300x200 (see rdW,rdH below)
        seg2 = new SIRectangle(0, 0, 0, 500, 0, ConfigManager.OBSERVED_REGION_CROSSHAIR_COLOR);
        if (!(Utils.osIsWindows() || Utils.osIsMacOS())) {
            this.observedRegion.setFilled(false);
        }
        this.rSpace.addGlyph(this.observedRegion);
        this.rSpace.addGlyph(seg1);
        this.rSpace.addGlyph(seg2);
        Glyph.stickToGlyph(seg1, this.observedRegion);
        Glyph.stickToGlyph(seg2, this.observedRegion);
        this.observedRegion.setSensitivity(false);
        this.tp = new ToolPalette(this);
        final Vector cameras = new Vector();
        cameras.add(this.mSpace.getCamera(0));
        cameras.add(this.mnSpace.getCamera(0));
        cameras.add(this.tp.getPaletteCamera());
        return cameras;
    }

    void createFrameView(final Vector cameras, final String vt, final JMenuBar jmb) {
        this.mainView =
            this.vsm.addFrameView(
                cameras,
                ConfigManager.MAIN_TITLE,
                vt,
                ConfigManager.mainViewW,
                ConfigManager.mainViewH,
                true,
                false,
                true,
                jmb);
        this.mainView.setLocation(ConfigManager.mainViewX, ConfigManager.mainViewY);
        this.mainView.getFrame().addComponentListener(this);
        this.gp = new ZGRGlassPane(this);
        ((JFrame) this.mainView.getFrame()).setGlassPane(this.gp);
    }

    JPanel createPanelView(final Vector cameras, final int w, final int h) {
        this.vsm.addPanelView(cameras, ConfigManager.MAIN_TITLE, View.STD_VIEW, w, h);
        this.mainView = this.vsm.getView(ConfigManager.MAIN_TITLE);
        this.gp = new ZGRGlassPane(this);
        return (JPanel) this.mainView.getPanel().getComponent();
    }

    public VirtualSpace getGraphSpace() {
        return this.mSpace;
    }

    public VirtualSpace getMenuSpace() {
        return this.mnSpace;
    }

    public View getView() {
        return this.mainView;
    }

    public Camera getGraphCamera() {
        return this.mainCamera;
    }

    public Camera getMenuCamera() {
        return this.mnSpace.getCamera(0);
    }

    public BaseEventHandler getViewListener() {
        return this.meh;
    }

    public ToolPalette getToolPalette() {
        return this.tp;
    }

    public VRectangle getMagWindow() {
        return this.magWindow;
    }

    public VRectangleOr getBoundingBox() {
        return this.boundingBox;
    }

    void parameterizeView(final BaseEventHandler eh) {
        this.paMngr = new PeriodicActionManager(this);
        this.mainView.setBackgroundColor(this.cfgMngr.backgroundColor);
        this.meh = eh;
        // same event handler handling all layers for now
        //XXX: TBD: refactor event handler code taking advantage of new one handler per layer functionality 
        this.mainView.setListener((ViewListener) eh, 0);
        this.mainView.setListener((ViewListener) eh, 1);
        //this.mainView.setListener((ViewListener) eh, 2);
        this.mainView.setNotifyCursorMoved(true);
        this.mainCamera.addListener(this);

        this.mainView.setVisible(true);
        this.mainView.getPanel().getComponent().addMouseMotionListener(this.paMngr);
        this.paMngr.start();
        this.mainView.setJava2DPainter(this.paMngr, Java2DPainter.AFTER_PORTALS);
        this.mainView.setJava2DPainter(this, Java2DPainter.FOREGROUND);

        activateDynaSpot(ConfigManager.DYNASPOT, false);
        this.mainView.getCursor().getDynaPicker().setDynaSpotColor(Color.RED);

        this.mainViewPanel = (JPanel) this.mainView.getPanel().getComponent();
        setAntialiasing(ConfigManager.ANTIALIASING);

        initDM();
        updatePanelSize();
        this.previousLocations = new Vector();

    }

    void activateDynaSpot(final boolean b, final boolean updatePrefs) {
        if (updatePrefs) {
            ConfigManager.DYNASPOT = b;
        }
        this.mainView.getCursor().getDynaPicker().activateDynaSpot(b);
    }

    void setConfigManager(final ConfigManager cm) {
        this.cfgMngr = cm;
    }

    void reset() {
        this.mSpace.removeAllGlyphs();
        this.mSpace.addGlyph(this.magWindow);
        this.mSpace.hide(this.magWindow);
        this.previousLocations.removeAllElements();
        this.highlightedEdges.removeAllElements();
        this.highlightedNodes.removeAllElements();
        this.originalEdgeColor.removeAllElements();
        this.originalNodeFillColor.removeAllElements();
        this.originalNodeBorderColor.removeAllElements();
    }

    void initDM() {
        this.magWindow = new VRectangle(0, 0, 0, 1, 1, GraphicsManager.DM_COLOR);
        this.magWindow.setFilled(false);
        this.magWindow.setBorderColor(GraphicsManager.DM_COLOR);
        this.mSpace.addGlyph(this.magWindow);
        this.mSpace.hide(this.magWindow);
    }

    /* Starting at version ? (somewhere between 1.16 and 2.8), GraphViz programs generate a polygon that bounds the entire graph.
    	* Attempt to identify it so that when clicking in what appears to be empty space (but is actually the bounding box),
    	* the view does not get unzoomed. Also prevent border highlighting when the cursor enters this bounding box.
    	*/
    void seekBoundingBox() {
        final Vector v = this.mSpace.getAllGlyphs();
        VRectangleOr largestRectangle = null;
        VRectangleOr r;
        int lri = -1; // largest rectangle's index
        // First identify the largest rectangle
        for (int i = 0; i < v.size(); i++) {
            if (v.elementAt(i) instanceof VRectangleOr) {
                r = (VRectangleOr) v.elementAt(i);
                if (largestRectangle == null || bigger(r, largestRectangle)) {
                    // first rectangle encountered in the list, or compare this rectangle to biggest rectangle at this time
                    largestRectangle = r;
                    lri = i;
                }
            }
        }
        if (lri == -1) {
            return;
        }
        // Then check that all other nodes are contained within that rectangle.
        //XXX: disabled check for objects lower in the stack as their seems to be some unidentified object that messes around
        // in some versions of GraphViz such as 2.14 - this workaround should be fairly safe until we identify that object
        //		for (int i=0;i<lri;i++){
        //			if (!containedIn((Glyph)v.elementAt(i), largestRectangle)){
        //				System.out.println(v.elementAt(i));
        //				//return;
        //			}
        //		}
        for (int i = lri + 1; i < v.size(); i++) {
            if (!containedIn((Glyph) v.elementAt(i), largestRectangle)) {
                return;
            }
        }
        // If they are, then it is very likely that the rectangle is a bounding box.
        this.boundingBox = largestRectangle;
        this.boundingBox.setVisible(false);
        this.boundingBox.setSensitivity(false);
    }

    boolean bigger(final VRectangleOr r1, final VRectangleOr r2) {
        // returns true if r1 bigger than r2
        return (r1.getWidth() * r1.getHeight() > r2.getWidth() * r2.getHeight());
    }

    boolean containedIn(final Glyph g, final VRectangle r) {
        if (g instanceof DPath || g instanceof VText) {
            return true;// don't take edges and text into accout, would be too costly (and would require one repaint for text)
        } else {// just get geometrical center for other glyphs ; this is an approximation, but it should work 
            return g.vx > r.vx - r.getWidth()
                && g.vx < r.vx + r.getWidth()
                && g.vy > r.vy - r.getHeight()
                && g.vy < r.vy + r.getHeight();
        }
    }

    /*antialias ON/OFF for views*/
    void setAntialiasing(final boolean b) {
        ConfigManager.ANTIALIASING = b;
        this.mainView.setAntialiasing(ConfigManager.ANTIALIASING);
    }

    /*-------------     Window resizing     -----------------*/

    void updatePanelSize() {
        this.tp.displayPalette(false);
        try {
            this.panelWidth = this.mainViewPanel.getWidth();
            this.panelHeight = this.mainViewPanel.getHeight();
            this.paMngr.requestToolPaletteRelocation();
            this.cfgMngr.notifyPlugins(Plugin.NOTIFY_PLUGIN_GUI_VIEW_RESIZED);
        } catch (final NullPointerException ex) {
        }
    }

    /*----------  Reveal graph (after loading) --------------*/

    void reveal() {
        final Camera c = this.mSpace.getCamera(0);
        final Location l = this.mainView.getGlobalView(c);
        c.vx = l.vx;
        c.vy = l.vy;
        c.setAltitude(l.alt - c.getFocal());
        rememberLocation(this.mSpace.getCamera(0).getLocation());
        Transitions.fadeIn(this.mainView, 500);
        getGlobalView();
    }

    /*-------------     Navigation              -------------*/

    public void getGlobalView() {
        final Location l = this.mainView.getGlobalView(this.mSpace.getCamera(0), ConfigManager.ANIM_MOVE_LENGTH);
        rememberLocation(this.mSpace.getCamera(0).getLocation());
    }

    /*higher view (multiply altitude by altitudeFactor)*/
    public void getHigherView() {
        final Camera c = this.mainView.getCameraNumber(0);
        rememberLocation(c.getLocation());
        final Float alt = new Float(c.getAltitude() + c.getFocal());
        final Animation a =
            this.animator.getAnimationFactory().createCameraAltAnim(
                ConfigManager.ANIM_MOVE_LENGTH,
                c,
                alt,
                true,
                SlowInSlowOutInterpolator.getInstance(),
                null);
        this.animator.startAnimation(a, false);
    }

    /*higher view (multiply altitude by altitudeFactor)*/
    public void getLowerView() {
        final Camera c = this.mainView.getCameraNumber(0);
        rememberLocation(c.getLocation());
        final Float alt = new Float(-(c.getAltitude() + c.getFocal()) / 2.0f);
        final Animation a =
            this.animator.getAnimationFactory().createCameraAltAnim(
                ConfigManager.ANIM_MOVE_LENGTH,
                c,
                alt,
                true,
                SlowInSlowOutInterpolator.getInstance(),
                null);
        this.animator.startAnimation(a, false);
    }

    /* Direction should be one of GraphicsManager.MOVE_* */
    public void translateView(final short direction) {
        final Camera c = this.mainView.getCameraNumber(0);
        rememberLocation(c.getLocation());
        Point2D.Double trans;
        final double[] rb = this.mainView.getVisibleRegion(c);
        if (direction == MOVE_UP) {
            final double qt = (rb[1] - rb[3]) / 2.4;
            trans = new Point2D.Double(0, qt);
        } else if (direction == MOVE_DOWN) {
            final double qt = (rb[3] - rb[1]) / 2.4;
            trans = new Point2D.Double(0, qt);
        } else if (direction == MOVE_RIGHT) {
            final double qt = (rb[2] - rb[0]) / 2.4;
            trans = new Point2D.Double(qt, 0);
        } else if (direction == MOVE_LEFT) {
            final double qt = (rb[0] - rb[2]) / 2.4;
            trans = new Point2D.Double(qt, 0);
        } else if (direction == MOVE_UP_LEFT) {
            final double qt = (rb[3] - rb[1]) / 2.4;
            final double qt2 = (rb[2] - rb[0]) / 2.4;
            trans = new Point2D.Double(qt, qt2);
        } else if (direction == MOVE_UP_RIGHT) {
            final double qt = (rb[1] - rb[3]) / 2.4;
            final double qt2 = (rb[2] - rb[0]) / 2.4;
            trans = new Point2D.Double(qt, qt2);
        } else if (direction == MOVE_DOWN_RIGHT) {
            final double qt = (rb[1] - rb[3]) / 2.4;
            final double qt2 = (rb[0] - rb[2]) / 2.4;
            trans = new Point2D.Double(qt, qt2);
        } else {
            //direction==DOWN_LEFT
            final double qt = (rb[3] - rb[1]) / 2.4;
            final double qt2 = (rb[0] - rb[2]) / 2.4;
            trans = new Point2D.Double(qt, qt2);
        }
        final Animation a =
            this.animator.getAnimationFactory().createCameraTranslation(
                ConfigManager.ANIM_MOVE_LENGTH,
                c,
                trans,
                true,
                SlowInSlowOutInterpolator.getInstance(),
                null);
        this.animator.startAnimation(a, false);
    }

    public void rememberLocation(final Location l) {
        if (this.previousLocations.size() >= MAX_PREV_LOC) {
            // as a result of release/click being undifferentiated)
            this.previousLocations.removeElementAt(0);
        }
        if (this.previousLocations.size() > 0) {
            if (!Location.equals((Location) this.previousLocations.lastElement(), l)) {
                this.previousLocations.add(l);
            }
        } else {
            this.previousLocations.add(l);
        }
    }

    public void moveBack() {
        if (this.previousLocations.size() > 0) {
            goTo((Location) this.previousLocations.lastElement());
        }
    }

    public void goTo(final Location l) {
        final Vector animParams = Location.getDifference(this.mSpace.getCamera(0).getLocation(), l);
        final Animation at =
            this.animator.getAnimationFactory().createCameraTranslation(
                ConfigManager.ANIM_MOVE_LENGTH,
                this.mSpace.getCamera(0),
                (Point2D.Double) animParams.elementAt(1),
                true,
                SlowInSlowOutInterpolator.getInstance(),
                null);
        final Animation aa =
            this.animator.getAnimationFactory().createCameraAltAnim(
                ConfigManager.ANIM_MOVE_LENGTH,
                this.mSpace.getCamera(0),
                (Double) animParams.elementAt(0),
                true,
                SlowInSlowOutInterpolator.getInstance(),
                null);
        this.animator.startAnimation(at, false);
        this.animator.startAnimation(aa, false);
    }

    /*show/hide radar view*/
    public void showRadarView(final boolean b) {
        if (b) {
            if (this.rView == null) {
                final Vector cameras = new Vector();
                cameras.add(this.mSpace.getCamera(1));
                cameras.add(this.rSpace.getCamera(0));
                this.vsm.addFrameView(
                    cameras,
                    RADAR_VIEW_NAME,
                    View.STD_VIEW,
                    ConfigManager.rdW,
                    ConfigManager.rdH,
                    true);
                this.reh = new RadarEvtHdlr(this);
                this.rView = this.vsm.getView(RADAR_VIEW_NAME);
                this.rView.getFrame().addComponentListener(this);
                this.rView.setBackgroundColor(this.cfgMngr.backgroundColor);
                // same event handler handling all layers for now
                //XXX: TBD: refactor event handler code taking advantage of new one-handler-per-layer functionality 
                this.rView.setListener(this.reh, 0);
                this.rView.setListener(this.reh, 1);
                this.rView.setResizable(true);
                this.rView.setActiveLayer(1);
                this.rView.setCursorIcon(java.awt.Cursor.MOVE_CURSOR);
                this.mSpace.getCamera(1).setLocation(this.rView.getGlobalView(this.mSpace.getCamera(1)));

                //rView.getGlobalView(mSpace.getCamera(1),100);
                // give null arguments because the method does not really care
                cameraMoved(null, null, 0);
                this.vsm.repaint();
            } else {
                ((JFrame) this.rView.getFrame()).toFront();
            }
        }
    }

    @Override
    public void cameraMoved(final Camera cam, final Point2D.Double coord, final double alt) {
        if (this.rView != null) {
            final Camera c0 = this.mSpace.getCamera(1);
            final Camera c1 = this.rSpace.getCamera(0);
            c1.vx = c0.vx;
            c1.vy = c0.vy;
            c1.focal = c0.focal;
            c1.altitude = c0.altitude;
            final double[] wnes = this.mainView.getVisibleRegion(this.mSpace.getCamera(0));
            this.observedRegion.moveTo((wnes[0] + wnes[2]) / 2, (wnes[3] + wnes[1]) / 2);
            this.observedRegion.setWidth((wnes[2] - wnes[0]));
            this.observedRegion.setHeight((wnes[1] - wnes[3]));
        }
        this.vsm.repaint();
    }

    public void updateMainViewFromRadar() {
        final Camera c0 = this.mSpace.getCamera(0);
        c0.vx = this.observedRegion.vx;
        c0.vy = this.observedRegion.vy;
        this.vsm.repaint();
    }

    //void centerRadarView(){
    //if (rView != null){
    //    rView.getGlobalView(mSpace.getCamera(1),ConfigManager.ANIM_MOVE_LENGTH);
    //    cameraMoved(null, null, 0);
    //}
    //}

    /*--------------------------- Lens management --------------------------*/

    public void setLens(final int t) {
        this.lensType = t;
    }

    public int getLensType() {
        return this.lensType;
    }

    public void moveLens(final int x, final int y, final long absTime) {
        if (this.fLens != null) {
            // dealing with a fading lens
            this.fLens.setAbsolutePosition(x, y, absTime);
        } else {
            // dealing with a probing lens
            this.lens.setAbsolutePosition(x, y);
        }
        this.vsm.repaint();
    }

    public Lens getLens() {
        return this.lens;
    }

    public void zoomInPhase1(final int x, final int y) {
        // create lens if it does not exist
        if (this.lens == null) {
            this.lens = this.mainView.setLens(getLensDefinition(x, y));
            this.lens.setBufferThreshold(1.5f);
        }
        final Animation a =
            this.animator.getAnimationFactory().createLensMagAnim(
                LENS_ANIM_TIME,
                (FixedSizeLens) this.lens,
                new Float(MAG_FACTOR - 1),
                true,
                IdentityInterpolator.getInstance(),
                null);
        this.animator.startAnimation(a, false);
        setLens(GraphicsManager.ZOOMIN_LENS);
    }

    public void zoomInPhase2(final double mx, final double my) {
        // compute camera animation parameters
        final double cameraAbsAlt = this.mainCamera.getAltitude() + this.mainCamera.getFocal();
        final double c2x = mx - INV_MAG_FACTOR * (mx - this.mainCamera.vx);
        final double c2y = my - INV_MAG_FACTOR * (my - this.mainCamera.vy);
        //Vector cadata = new Vector();
        // -(cameraAbsAlt)*(MAG_FACTOR-1)/MAG_FACTOR
        final Double deltAlt = new Double((cameraAbsAlt) * (1 - MAG_FACTOR) / MAG_FACTOR);
        if (cameraAbsAlt + deltAlt.floatValue() > FLOOR_ALTITUDE) {
            final Animation al =
                this.animator.getAnimationFactory().createLensMagAnim(
                    LENS_ANIM_TIME,
                    (FixedSizeLens) this.lens,
                    new Float(-MAG_FACTOR + 1),
                    true,
                    IdentityInterpolator.getInstance(),
                    new ZP2LensAction(this));
            final Animation at =
                this.animator.getAnimationFactory().createCameraTranslation(
                    LENS_ANIM_TIME,
                    this.mainCamera,
                    new Point2D.Double(c2x - this.mainCamera.vx, c2y - this.mainCamera.vy),
                    true,
                    IdentityInterpolator.getInstance(),
                    null);
            final Animation aa =
                this.animator.getAnimationFactory().createCameraAltAnim(
                    LENS_ANIM_TIME,
                    this.mainCamera,
                    deltAlt,
                    true,
                    IdentityInterpolator.getInstance(),
                    null);
            this.animator.startAnimation(al, false);
            this.animator.startAnimation(at, false);
            this.animator.startAnimation(aa, false);
        } else {
            final Double actualDeltAlt = new Double(FLOOR_ALTITUDE - cameraAbsAlt);
            final double ratio = actualDeltAlt.doubleValue() / deltAlt.doubleValue();
            final Animation al =
                this.animator.getAnimationFactory().createLensMagAnim(
                    LENS_ANIM_TIME,
                    (FixedSizeLens) this.lens,
                    new Float(-MAG_FACTOR + 1),
                    true,
                    IdentityInterpolator.getInstance(),
                    new ZP2LensAction(this));
            final Animation at =
                this.animator.getAnimationFactory().createCameraTranslation(
                    LENS_ANIM_TIME,
                    this.mainCamera,
                    new Point2D.Double((c2x - this.mainCamera.vx) * ratio, (c2y - this.mainCamera.vy) * ratio),
                    true,
                    IdentityInterpolator.getInstance(),
                    null);
            final Animation aa =
                this.animator.getAnimationFactory().createCameraAltAnim(
                    LENS_ANIM_TIME,
                    this.mainCamera,
                    actualDeltAlt,
                    true,
                    IdentityInterpolator.getInstance(),
                    null);
            this.animator.startAnimation(al, false);
            this.animator.startAnimation(at, false);
            this.animator.startAnimation(aa, false);
        }
    }

    public void zoomOutPhase1(final int x, final int y, final double mx, final double my) {
        // compute camera animation parameters
        final double cameraAbsAlt = this.mainCamera.getAltitude() + this.mainCamera.getFocal();
        final double c2x = mx - MAG_FACTOR * (mx - this.mainCamera.vx);
        final double c2y = my - MAG_FACTOR * (my - this.mainCamera.vy);
        // create lens if it does not exist
        if (this.lens == null) {
            this.lens = this.mainView.setLens(getLensDefinition(x, y));
            this.lens.setBufferThreshold(1.5f);
        }
        // animate lens and camera simultaneously
        final Animation al =
            this.animator.getAnimationFactory().createLensMagAnim(
                LENS_ANIM_TIME,
                (FixedSizeLens) this.lens,
                new Float(MAG_FACTOR - 1),
                true,
                IdentityInterpolator.getInstance(),
                null);
        final Animation at =
            this.animator.getAnimationFactory().createCameraTranslation(
                LENS_ANIM_TIME,
                this.mainCamera,
                new Point2D.Double(c2x - this.mainCamera.vx, c2y - this.mainCamera.vy),
                true,
                IdentityInterpolator.getInstance(),
                null);
        final Animation aa =
            this.animator.getAnimationFactory().createCameraAltAnim(
                LENS_ANIM_TIME,
                this.mainCamera,
                new Double(cameraAbsAlt * (MAG_FACTOR - 1)),
                true,
                IdentityInterpolator.getInstance(),
                null);
        this.animator.startAnimation(al, false);
        this.animator.startAnimation(at, false);
        this.animator.startAnimation(aa, false);
        setLens(GraphicsManager.ZOOMOUT_LENS);
    }

    public void zoomOutPhase2() {
        // make lens disappear (killing anim)
        final Animation a =
            this.animator.getAnimationFactory().createLensMagAnim(
                LENS_ANIM_TIME,
                (FixedSizeLens) this.lens,
                new Float(-MAG_FACTOR + 1),
                true,
                IdentityInterpolator.getInstance(),
                new ZP2LensAction(this));
        this.animator.startAnimation(a, false);
    }

    public void setMagFactor(final double m) {
        MAG_FACTOR = m;
        INV_MAG_FACTOR = 1 / MAG_FACTOR;
    }

    public synchronized void magnifyFocus(final double magOffset, final int zooming, final Camera ca) {
        synchronized (this.lens) {
            final double nmf = MAG_FACTOR + magOffset;
            if (nmf <= MAX_MAG_FACTOR && nmf > 1.0f) {
                setMagFactor(nmf);
                if (zooming == GraphicsManager.ZOOMOUT_LENS) {
                    /* if unzooming, we want to keep the focus point stable, and unzoom the context
                        this means that camera altitude must be adjusted to keep altitude + lens mag
                        factor constant in the lens focus region. The camera must also be translated
                        to keep the same region of the virtual space under the focus region */
                    final double a1 = this.mainCamera.getAltitude();
                    this.lens.setMaximumMagnification((float) nmf, true);
                    /* explanation for the altitude offset computation: the projected size of an object
                        in the focus region (in lens space) should remain the same before and after the
                        change of magnification factor. The size of an object is a function of the
                        projection coefficient (see any Glyph.projectForLens() method). This means that
                        the following equation must be true, where F is the camera's focal distance, a1
                        is the camera's altitude before the move and a2 is the camera altitude after the
                        move:
                    MAG_FACTOR * F / (F + a1) = MAG_FACTOR + magOffset * F / (F + a2)
                        From this we can get the altitude difference (a2 - a1)                       */
                    this.mainCamera.altitudeOffset((a1 + this.mainCamera.getFocal())
                        * magOffset
                        / (MAG_FACTOR - magOffset));
                    /* explanation for the X offset computation: the position in X of an object in the
                        focus region (lens space) should remain the same before and after the change of
                        magnification factor. This means that the following equation must be true (taken
                        by simplifying pc[i].lcx computation in a projectForLens() method):
                    (vx-(lensx1))*coef1 = (vx-(lensx2))*coef2
                        -- coef1 is actually MAG_FACTOR * F/(F+a1)
                        -- coef2 is actually (MAG_FACTOR + magOffset) * F/(F+a2)
                        -- lensx1 is actually camera.vx1 + ((F+a1)/F) * lens.lx
                        -- lensx2 is actually camera.vx2 + ((F+a2)/F) * lens.lx
                        Given that (MAG_FACTOR + magOffset) / (F+a2) = MAG_FACTOR / (F+a1)
                        we eventually have:
                    Xoffset = (a1 - a2) / F * lens.lx   (lens.lx being the position of the lens's center in
                        JPanel coordinates w.r.t the view's center - see Lens.java)
                        */
                    this.mainCamera.move((a1 - this.mainCamera.getAltitude()
                        / this.mainCamera.getFocal()
                        * this.lens.lx), -(a1 - this.mainCamera.getAltitude()
                        / this.mainCamera.getFocal()
                        * this.lens.ly));
                } else {
                    final Animation a =
                        this.animator.getAnimationFactory().createLensMagAnim(
                            WHEEL_ANIM_TIME,
                            (FixedSizeLens) this.lens,
                            new Float(magOffset),
                            true,
                            IdentityInterpolator.getInstance(),
                            null);
                    this.animator.startAnimation(a, false);
                }
            }
        }
    }

    Lens getLensDefinition(final int x, final int y) {
        Lens res;
        if (this.tp.isFadingLensNavMode()) {
            this.fLens = new LInfSCBLens(1.0f, 0.0f, 0.95f, 100, x - this.panelWidth / 2, y - this.panelHeight / 2);
            res = this.fLens;
        } else {
            // isProbingLensNavMode()
            res = new FSGaussianLens(1.0f, 100, 50, x - this.panelWidth / 2, y - this.panelHeight / 2);
            // unset any previous fading lens to make sure it gets garbage collected
            this.fLens = null;
        }
        return res;
    }

    /*-------------        DragMag        -----------------*/

    public Camera getDragMagCamera() {
        return this.dmCamera;
    }

    public DraggableCameraPortal getDragMagPortal() {
        return this.dmPortal;
    }

    public void triggerDM(final int x, final int y, final PortalListener pl) {
        if (this.dmPortal != null) {
            // portal is active, destroy it
            killDM();
        } else {
            // portal not active, create it
            createDM(x, y, pl);
        }
    }

    void createDM(final int x, final int y, final PortalListener pl) {
        this.dmPortal =
            new DraggableCameraPortal(
                x,
                y,
                GraphicsManager.DM_PORTAL_WIDTH,
                GraphicsManager.DM_PORTAL_HEIGHT,
                this.dmCamera);
        this.dmPortal.setPortalListener(pl);
        this.dmPortal.setBackgroundColor(this.mainView.getBackgroundColor());
        this.vsm.addPortal(this.dmPortal, this.mainView);
        this.dmPortal.setBorder(GraphicsManager.DM_COLOR);
        final Location l = this.dmPortal.getSeamlessView(this.mainCamera);
        this.dmCamera.moveTo(l.vx, l.vy);
        this.dmCamera
            .setAltitude(((this.mainCamera.getAltitude() + this.mainCamera.getFocal()) / (DEFAULT_MAG_FACTOR) - this.mainCamera
                .getFocal()));
        updateMagWindow();
        final int w =
            (int) Math.round(this.magWindow.getWidth()
                * this.mainCamera.getFocal()
                / ((float) (this.mainCamera.getFocal() + this.mainCamera.getAltitude())));
        final int h =
            (int) Math.round(this.magWindow.getHeight()
                * this.mainCamera.getFocal()
                / ((float) (this.mainCamera.getFocal() + this.mainCamera.getAltitude())));
        this.dmPortal.sizeTo(w, h);
        this.mSpace.onTop(this.magWindow);
        this.mSpace.show(this.magWindow);
        this.paintLinks = true;
        //animator.createPortalAnimation(GraphicsManager.DM_PORTAL_ANIM_TIME, AnimManager.PT_SZ_TRANS_LIN, data, dmPortal.getID(), null);
        final Animation as =
            this.animator.getAnimationFactory().createPortalSizeAnim(
                GraphicsManager.DM_PORTAL_ANIM_TIME,
                this.dmPortal,
                GraphicsManager.DM_PORTAL_WIDTH - w,
                GraphicsManager.DM_PORTAL_HEIGHT - h,
                true,
                IdentityInterpolator.getInstance(),
                null);
        final Animation at =
            this.animator.getAnimationFactory().createPortalTranslation(
                GraphicsManager.DM_PORTAL_ANIM_TIME,
                this.dmPortal,
                new Point(
                    GraphicsManager.DM_PORTAL_INITIAL_X_OFFSET - w / 2,
                    GraphicsManager.DM_PORTAL_INITIAL_Y_OFFSET - h / 2),
                true,
                IdentityInterpolator.getInstance(),
                null);
        this.animator.startAnimation(as, false);
        this.animator.startAnimation(at, false);
    }

    void killDM() {
        if (this.dmPortal != null) {
            this.vsm.destroyPortal(this.dmPortal);
            this.dmPortal = null;
            this.mSpace.hide(this.magWindow);
            this.paintLinks = false;
        }
        this.meh.resetDragMagInteraction();
    }

    double[] dmwnes = new double[4];

    public void updateMagWindow() {
        if (this.dmPortal == null) {
            return;
        }
        this.dmPortal.getVisibleRegion(this.dmwnes);
        this.magWindow.moveTo(this.dmCamera.vx, this.dmCamera.vy);
        this.magWindow.setWidth((this.dmwnes[2] - this.dmwnes[0]) + 1);
        this.magWindow.setHeight((this.dmwnes[1] - this.dmwnes[3]) + 1);
    }

    public void updateZoomWindow() {
        this.dmCamera.moveTo(this.magWindow.vx, this.magWindow.vy);
    }

    /*Java2DPainter interface*/
    @Override
    public void paint(final Graphics2D g2d, final int viewWidth, final int viewHeight) {
        if (this.paintLinks) {
            final double coef = this.mainCamera.focal / (this.mainCamera.focal + this.mainCamera.altitude);
            final int magWindowX = (int) Math.round((viewWidth / 2) + (this.magWindow.vx - this.mainCamera.vx) * coef);
            final int magWindowY = (int) Math.round((viewHeight / 2) - (this.magWindow.vy - this.mainCamera.vy) * coef);
            final int magWindowW = (int) Math.round(this.magWindow.getWidth() * coef / 2);
            final int magWindowH = (int) Math.round(this.magWindow.getHeight() * coef / 2);
            g2d.setColor(GraphicsManager.DM_COLOR);
            g2d.drawLine(magWindowX - magWindowW, magWindowY - magWindowH, this.dmPortal.x, this.dmPortal.y);
            g2d.drawLine(
                magWindowX + magWindowW,
                magWindowY - magWindowH,
                this.dmPortal.x + this.dmPortal.w,
                this.dmPortal.y);
            g2d.drawLine(magWindowX - magWindowW, magWindowY + magWindowH, this.dmPortal.x, this.dmPortal.y
                + this.dmPortal.h);
            g2d.drawLine(
                magWindowX + magWindowW,
                magWindowY + magWindowH,
                this.dmPortal.x + this.dmPortal.w,
                this.dmPortal.y + this.dmPortal.h);
        }
        g2d.setFont(ConfigManager.SMALL_FONT);
        if (this.tp.isHighlightMode()) {
            g2d.drawString("HL = " + this.highlightLevel, viewWidth - 50, 20);
        }
    }

    /* ---------- search -----------*/

    Glyph highlightedLabel;
    Color originalHighlightedLabelColor;

    /*given a string, centers on a VText with this string in it*/
    void search(final String s, final int direction) {
        if (s.length() > 0) {
            if (!s.toLowerCase().equals(this.lastSearchedString)) {//searching a new string - reinitialize everything
                resetSearch(s);
                final Glyph[] gl = this.mSpace.getVisibleGlyphsList();
                for (final Glyph element : gl) {
                    if (element instanceof VText) {
                        if ((((VText) element).getText() != null)
                            && (((VText) element).getText().toLowerCase().indexOf(this.lastSearchedString) != -1))
                        {
                            this.matchingList.add(element);
                        }
                    }
                }
            }
            final int matchSize = this.matchingList.size();
            if (matchSize > 0) {
                //get prev/next entry in the list of matching elements
                this.searchIndex = this.searchIndex + direction;
                if (this.searchIndex < 0) {// if reached start/end of list, go to end/start (loop)
                    this.searchIndex = matchSize - 1;
                } else if (this.searchIndex >= matchSize) {
                    this.searchIndex = 0;
                }
                if (matchSize > 1) {
                    this.zapp.setStatusBarText(AppletUtils.rankString(this.searchIndex + 1)
                        + " of "
                        + matchSize
                        + " matches");
                } else {
                    this.zapp.setStatusBarText("1 match");
                }
                //center on the entity
                final Glyph g = (Glyph) this.matchingList.elementAt(this.searchIndex);
                // g is the lastMatchingEntity
                this.mainView.centerOnGlyph(
                    g,
                    this.mSpace.getCamera(0),
                    ConfigManager.ANIM_MOVE_LENGTH,
                    true,
                    ConfigManager.MAG_FACTOR * 1.5f);
                highlight(g);
                this.vsm.repaint();
            } else {
                this.zapp.setStatusBarText("No match");
            }
        }
    }

    /*reset the search variables after it is finished*/
    void resetSearch(final String s) {
        this.searchIndex = -1;
        this.lastSearchedString = s.toLowerCase();
        this.matchingList.removeAllElements();
        if (this.highlightedLabel != null) {
            this.highlightedLabel.setColor(this.originalHighlightedLabelColor);
            this.highlightedLabel = null;
        }
    }

    /* color the label found by search */
    void highlight(final Glyph g) {
        // de-highlight previous label (if any)
        if (this.highlightedLabel != null) {
            this.highlightedLabel.setColor(this.originalHighlightedLabelColor);
        }
        this.highlightedLabel = g;
        this.originalHighlightedLabelColor = this.highlightedLabel.getColor();
        this.highlightedLabel.setColor(ConfigManager.HIGHLIGHT_COLOR);
    }

    /* -------- Progress bar on glass pane -----------*/

    /* -------------- Font management ----------------*/

    void assignFontToGraph() {
        final Font f =
            fr.inria.zvtm.widgets.FontDialog
                .getFontDialog((JFrame) this.mainView.getFrame(), ConfigManager.defaultFont);
        if (f != null) {
            ConfigManager.defaultFont = f;
            final Vector glyphs = this.mSpace.getAllGlyphs();
            Object g;
            for (int i = 0; i < glyphs.size(); i++) {
                g = glyphs.elementAt(i);
                if (g instanceof VText) {
                    ((VText) g).setFont(null);
                }
            }
            VText.setMainFont(ConfigManager.defaultFont);
        }
    }

    @Override
    public void componentResized(final ComponentEvent e) {
        if (e.getSource() == this.mainView.getFrame()) {
            updatePanelSize();
            //update rectangle showing observed region in radar view when main view's aspect ratio changes
            cameraMoved(null, null, 0);
            //update SD_ZOOM_THRESHOLD
            final Dimension sz = this.mainView.getFrame().getSize();
            this.cfgMngr.setSDZoomThreshold(0.3 * Math.sqrt(Math.pow(sz.width, 2) + Math.pow(sz.height, 2)));
        } else if (e.getSource() == this.rView.getFrame()) {
            this.mSpace.getCamera(1).setLocation(this.rView.getGlobalView(this.mSpace.getCamera(1)));
            cameraMoved(null, null, 0);
            this.vsm.repaint();
        }
    }

    @Override
    public void componentHidden(final ComponentEvent e) {
    }

    @Override
    public void componentMoved(final ComponentEvent e) {
    }

    @Override
    public void componentShown(final ComponentEvent e) {
    }

    /* ------------- Logical structure ----------------- */
    void buildLogicalStructure() {
        // clone the structure as we are about to remove elements from it for convenience (it is supposed to be read-only)
        Vector glyphs = (Vector) this.mSpace.getAllGlyphs().clone();
        glyphs.remove(this.magWindow);
        glyphs.remove(this.boundingBox);
        this.lstruct = LogicalStructure.build(glyphs, this.mSpace);
        if (this.lstruct == null) {
            // building the logical structure failed
            this.mainView.setStatusBarText(Messages.FAILED_TO_BUILD_LOGICAL_STRUCT);
        }
        /* take care of converting the owner of glyphs that were not processed as structural elements
        (which should have remained Metadata instances). Convert these old owners into LElem instances
        (as this is what ZGRViewer now expects to get when calling Glyph.getOwner()) */
        glyphs = this.mSpace.getAllGlyphs(); // not cloning here because we are just reading the structure
        Glyph g;
        for (int i = 0; i < glyphs.size(); i++) {
            g = (Glyph) glyphs.elementAt(i);
            if (g.getOwner() != null && g.getOwner() instanceof Metadata) {
                g.setOwner(new LElem((Metadata) g.getOwner()));
            }
        }
        this.cfgMngr.notifyPlugins(Plugin.NOTIFY_PLUGIN_LOGICAL_STRUCTURE_CHANGED);
    }

    /* ------------- Graph editing ---------------- */

    static final float[] TRIANGLE_VERTICES = {1f, 1f, 1f };

    public LEdge addEdge(final Glyph sn, final Glyph en, final String title, final boolean directed) {
        final LNode s = LogicalStructure.getNode(sn);
        final LNode e = LogicalStructure.getNode(en);
        if (s != null && e != null) {
            return addEdge(s, e, title, directed);
        } else {
            return null;
        }
    }

    public LEdge addEdge(final LNode sn, final LNode en, final String title, final boolean directed) {
        final ClosedShape sns = sn.getShape();
        final ClosedShape ens = en.getShape();
        // find endpoints for the edge by computing intersections between segment
        // linking both node shapes' center and each of those shape's boundary
        final BasicStroke st = new BasicStroke(.2f);
        final Shape snsh = st.createStrokedShape(sns.getJava2DShape());
        final Shape ensh = st.createStrokedShape(ens.getJava2DShape());
        final Area edgeseg = new Area(st.createStrokedShape(new Line2D.Double(sns.vx, sns.vy, ens.vx, ens.vy)));
        final Area intersectStart = new Area(snsh);
        intersectStart.intersect(edgeseg);
        final Area intersectEnd = new Area(ensh);
        intersectEnd.intersect(edgeseg);
        final Rectangle2D spc = intersectStart.getBounds2D();
        final Rectangle2D epc = intersectEnd.getBounds2D();
        // compute control point coordinates for both quadratic curves
        // control point of 1st quadratic curve
        final Point2D.Double cpqd1 =
            new Point2D.Double(
                (3 * spc.getCenterX() + epc.getCenterX()) / 4d,
                (3 * spc.getCenterY() + epc.getCenterY()) / 4d);
        // point of junction between both quadratic curves
        final Point2D.Double jp =
            new Point2D.Double((spc.getCenterX() + epc.getCenterX()) / 2d, (spc.getCenterY() + epc.getCenterY()) / 2d);
        // control point of 2nd quadratic curve
        final Point2D.Double cpqd2 =
            new Point2D.Double(
                (spc.getCenterX() + 3 * epc.getCenterX()) / 4d,
                (spc.getCenterY() + 3 * epc.getCenterY()) / 4d);
        // instantiate all glyphs that form the edge
        final DPath spline = new DPath(spc.getCenterX(), spc.getCenterY(), 0, Color.BLACK);
        spline.addQdCurve(jp.x, jp.y, cpqd1.x, cpqd1.y, true);
        spline.addQdCurve(epc.getCenterX(), epc.getCenterY(), cpqd2.x, cpqd2.y, true);
        final double theta = Math.atan2(epc.getCenterY() - spc.getCenterY(), epc.getCenterX() - spc.getCenterX());
        final VShape arrowHead =
            new VShape(
                epc.getCenterX(),
                epc.getCenterY(),
                0,
                ens.getSize() / 10.0,
                TRIANGLE_VERTICES,
                Color.BLACK,
                theta);
        // put them in virtual space
        SwingUtilities.invokeLater(new Runnable() {
            @Override
            public void run() {
                GraphicsManager.this.mSpace.addGlyph(spline);
                GraphicsManager.this.mSpace.addGlyph(arrowHead);
            }
        });
        // create the LEdge
        final Vector<Glyph> glyphs = new Vector(2);
        glyphs.add(spline);
        glyphs.add(arrowHead);
        final LEdge res = new LEdge(glyphs);
        res.setDirected(directed);
        res.setTail(sn);
        res.setHead(en);
        this.lstruct.addEdge(res);
        return res;
    }

    public void removeEdge(final LEdge e) {
        this.lstruct.removeEdge(e);
        SwingUtilities.invokeLater(new Runnable() {
            @Override
            public void run() {
                for (final Glyph g : e.getGlyphs()) {
                    GraphicsManager.this.mSpace.removeGlyph(g, false);
                }
                VirtualSpaceManager.INSTANCE.repaint();
            }
        });
    }

    /* ------------- Highlighting ----------------- */

    static final BasicStroke HIGHLIGHT_STROKE = new BasicStroke(2.0f);

    Vector highlightedEdges = new Vector();
    Vector originalEdgeColor = new Vector();
    Vector originalEdgeBorderColor = new Vector();
    Vector originalEdgeStroke = new Vector();

    Vector highlightedNodes = new Vector();
    Vector originalNodeBorderColor = new Vector();
    Vector originalNodeFillColor = new Vector();
    Vector originalNodeStroke = new Vector();

    // currently selected level of highlighting neighboring nodes
    int highlightLevel;
    // currently selected direction of highlighting neighboring -1/UNDIRECTED/INCOMING/OUTGOING
    int highlightDir;
    // keep already highlighted nodes. This is for performance
    HashMap highlightNodesDone = new HashMap(1000);
    // timeout for overly deep highlighting in seconds
    final int highlightTimeStampTimeout = 5;
    Calendar highlightTimeStamp = Calendar.getInstance();
    // nodes fixated for highlighting 
    Vector highlightNodes = new Vector();
    // number of levels for the fixated node at according pos of highlightNodes
    Vector highlightLevels = new Vector();
    // direction of highlighting for the fixated node at according pos of highlightNodes
    Vector highlightDirections = new Vector();

    public void highlightElement(
        Glyph g,
        final Camera cam,
        final VCursor cursor,
        final boolean highlight,
        final int moreLess,
        final boolean fixNode,
        final int dir)
    {
        Object o = null;
        if (g != null && g != this.boundingBox) {
            // clicked inside a node
            o = g.getOwner();
        } else {
            // if cursor was not in a shape, try to detect a label or an edge           
            final Vector otherGlyphs = cursor.getPicker().getIntersectingGlyphs(cam);
            if (otherGlyphs != null && otherGlyphs.size() > 0) {
                g = (Glyph) otherGlyphs.firstElement();
                if (g.getOwner() != null) {
                    o = g.getOwner();
                }
            } else {
                // could not detect anything, consider that user clicked on background
                // unhighlight anything that could have been highlighted
                unhighlightAll();
            }
        }
        if (o != null) {
            if (fixNode) {
                final int indexNode = this.highlightNodes.indexOf(o);
                if (indexNode >= 0) {
                    this.highlightNodes.remove(indexNode);
                    this.highlightLevels.remove(indexNode);
                    this.highlightDirections.remove(indexNode);
                    unhighlightAll();
                } else {
                    this.highlightNodes.add(o);
                    this.highlightLevels.add(new Integer(this.highlightLevel));
                    this.highlightDirections.add(new Integer(this.highlightDir));
                }
                return;
            } else {
                if (moreLess == 0) {
                    this.highlightLevel = 1;
                    this.highlightDir = -1;
                } else {
                    this.highlightLevel = Math.max(this.highlightLevel + moreLess, 0);
                    this.highlightDir = dir;
                }
            }
            unhighlightAll();

            if (o instanceof LNode) {
                highlightNodeInOut((LNode) o, highlight, this.highlightLevel, dir);
            } else if (o instanceof LEdge) {
                highlightEdge((LEdge) o, highlight);
            }
        }
    }

    public void highlightPermanent() {
        for (int indexNode = 0; indexNode < this.highlightNodes.size(); indexNode++) {
            final LNode n = (LNode) this.highlightNodes.get(indexNode);
            final Integer level = (Integer) this.highlightLevels.get(indexNode);
            final Integer dir = (Integer) this.highlightDirections.get(indexNode);
            highlightNodeInOut(n, true, level.intValue(), dir.intValue());
        }
    }

    synchronized
        void
        highlightNodeInOut(final LNode n, final boolean highlight, final int highlightLevel, final int dir)
    {
        this.highlightNodesDone.clear();
        this.highlightTimeStamp = Calendar.getInstance();
        this.highlightTimeStamp.add(Calendar.SECOND, this.highlightTimeStampTimeout);
        if (dir == LEdge.UNDIRECTED) {
            highlightNode(n, highlight, LEdge.UNDIRECTED, highlightLevel);
        } else {
            if (dir != LEdge.OUTGOING) {
                highlightNode(n, highlight, LEdge.INCOMING, highlightLevel);
            }
            this.highlightNodesDone.clear();
            this.highlightTimeStamp = Calendar.getInstance();
            this.highlightTimeStamp.add(Calendar.SECOND, this.highlightTimeStampTimeout);
            if (dir != LEdge.INCOMING) {
                highlightNode(n, highlight, LEdge.OUTGOING, highlightLevel);
            }
        }
    }

    synchronized void highlightNode(final LNode n, final boolean highlight, final short dir, final int levelCnt) {
        if (highlight) {
            if (Calendar.getInstance().after(this.highlightTimeStamp)) {
                unhighlightAll();
                return;
            }
            final Integer nodeLevel = ((Integer) this.highlightNodesDone.get(n));
            if (nodeLevel != null && nodeLevel.intValue() >= levelCnt) {
                return;
            }
            this.highlightNodesDone.put(n, new Integer(levelCnt));
            Glyph g;
            final Glyph[] gs;
            // highlight node itself
            for (final Glyph glyph : n.glyphs) {
                g = glyph;
                if (this.highlightedNodes.contains(g)) {
                    continue;
                }
                this.highlightedNodes.add(g);
                highlightNodeGlyph(g);
            }
            if (levelCnt <= 0) {
                return;
            }
            // for all edges linked to this node
            for (int i = 0; i < n.edges.length; i++) {
                // highlight edge itself
                if (dir != LEdge.UNDIRECTED && dir != n.edgeDirections[i] && n.edges[i].isDirected()) {
                    continue;
                }
                for (final Glyph glyph : n.edges[i].glyphs) {
                    g = glyph;
                    /* prevent elements from being processed more than once
                       this can happen when a node links to itself, and
                       this makes the highlighting mechanism think that
                       the arc's original color is the selection color */
                    if (this.highlightedEdges.contains(g)) {
                        continue;
                    }
                    this.highlightedEdges.add(g);
                    highlightEdgeGlyph(g);
                }
                // highlight node at other end of arc
                highlightNode(n.edges[i].getOtherEnd(n), highlight, dir, levelCnt - 1);
            }
        } else {
            unhighlightAll();
        }
    }

    synchronized void highlightEdge(final LEdge e, final boolean highlight) {
        if (highlight) {
            Glyph g;
            // highlight edge itself
            for (final Glyph glyph : e.glyphs) {
                g = glyph;
                if (this.highlightedEdges.contains(g)) {
                    continue;
                }
                this.highlightedEdges.add(g);
                highlightEdgeGlyph(g);
            }
            // highlight tail
            for (final Glyph glyph : e.tail.glyphs) {
                g = glyph;
                /* prevent elements from being processed more than once
                   this can happen when a node links to itself, and
                   this makes the highlighting mechanism think that
                   the arc's original color is the selection color */
                if (this.highlightedNodes.contains(g)) {
                    continue;
                }
                this.highlightedNodes.add(g);
                highlightNodeGlyph(g);
            }
            // highlight head
            for (final Glyph glyph : e.head.glyphs) {
                g = glyph;
                /* prevent elements from being processed more than once
                   this can happen when a node links to itself, and
                   this makes the highlighting mechanism think that
                   the arc's original color is the selection color */
                if (this.highlightedNodes.contains(g)) {
                    continue;
                }
                this.highlightedNodes.add(g);
                highlightNodeGlyph(g);
            }
        } else {
            unhighlightAll();
        }
    }

    void highlightNodeGlyph(final Glyph g) {
        this.originalNodeStroke.add(g.getStroke());
        if (g instanceof ClosedShape) {
            // keep both originalXXXColor vectors at the same length/indexes for a given glyph
            this.originalNodeFillColor.add(null);
            this.originalNodeBorderColor.add(((ClosedShape) g).getDefaultBorderColor());
            if (g.isBorderDrawn()) {
                g.setBorderColor(ConfigManager.HIGHLIGHT_COLOR);
            }
        } else {
            // keep both originalXXXColor vectors at the same length/indexes for a given glyph
            this.originalNodeFillColor.add(null);
            this.originalNodeBorderColor.add(g.getDefaultColor());
            g.setColor(ConfigManager.HIGHLIGHT_COLOR);
        }
        g.setStroke(HIGHLIGHT_STROKE);
    }

    void highlightEdgeGlyph(final Glyph g) {
        this.originalEdgeColor.add(g.getDefaultColor());
        this.originalEdgeStroke.add(g.getStroke());
        if (g instanceof ClosedShape) {
            this.originalEdgeBorderColor.add(((ClosedShape) g).getDefaultBorderColor());
            if (g.isFilled()) {
                // use border color to fill arrow head shape
                g.setColor(ConfigManager.HIGHLIGHT_COLOR);
            }
            if (g.isBorderDrawn()) {
                g.setBorderColor(ConfigManager.HIGHLIGHT_COLOR);
            }
        } else {
            this.originalEdgeBorderColor.add(Color.BLACK);
            g.setColor(ConfigManager.HIGHLIGHT_COLOR);
        }
        g.setStroke(HIGHLIGHT_STROKE);
    }

    public void unhighlightAll() {
        unhighlightAllEdges();
        unhighlightAllNodes();
        highlightPermanent();
    }

    void unhighlightAllNodes() {
        Glyph g;
        for (int i = 0; i < this.highlightedNodes.size(); i++) {
            g = (Glyph) this.highlightedNodes.elementAt(i);
            if (g instanceof ClosedShape) {
                if (g.isBorderDrawn()) {
                    g.setBorderColor((Color) this.originalNodeBorderColor.elementAt(i));
                }
            } else {
                g.setColor((Color) this.originalNodeBorderColor.elementAt(i));
            }
            g.setStroke((BasicStroke) this.originalNodeStroke.elementAt(i));
        }
        this.highlightedNodes.removeAllElements();
        this.originalNodeBorderColor.removeAllElements();
        this.originalNodeFillColor.removeAllElements();
        this.originalNodeStroke.removeAllElements();
    }

    void unhighlightAllEdges() {
        Glyph g;
        for (int i = 0; i < this.highlightedEdges.size(); i++) {
            g = (Glyph) this.highlightedEdges.elementAt(i);
            if (g instanceof ClosedShape) {
                if (g.isFilled()) {
                    g.setColor((Color) this.originalEdgeColor.elementAt(i));
                }
                if (g.isBorderDrawn()) {
                    g.setBorderColor((Color) this.originalEdgeBorderColor.elementAt(i));
                }
            } else {
                g.setColor((Color) this.originalEdgeColor.elementAt(i));
            }
            g.setStroke((BasicStroke) this.originalEdgeStroke.elementAt(i));
        }
        this.highlightedEdges.removeAllElements();
        this.originalEdgeColor.removeAllElements();
        this.originalEdgeBorderColor.removeAllElements();
        this.originalEdgeStroke.removeAllElements();
    }

    /* -------------- Bring and Go mode (previously called Fresnel mode) -------------------- */

    static final int BRING_ANIM_DURATION = 300;
    static final double BRING_DISTANCE_FACTOR = 1.5;

    static final float FADED_ELEMENTS_TRANSLUCENCY = 0.1f;
    static final float SECOND_STEP_TRANSLUCENCY = 0.3f;

    boolean isBringingAndGoing = false;

    public boolean isBringingAndGoing() {
        return this.isBringingAndGoing;
    }

    Vector broughtElements = new Vector();

    Vector elementsToFade;
    float[] alphaOfElementsToFade;

    void enterBringAndGoMode() {
    }

    void exitBringAndGoMode() {
    }

    void startBringAndGo(final Glyph g) {
        this.isBringingAndGoing = true;
        final LNode n = LogicalStructure.getNode(g);
        if (n == null) {
            return;
        }
        // prepare list of elements to fade out
        // (it contains too many elements at this point, but these will be
        // progressively removed from the list as we process connected nodes)
        this.elementsToFade = (Vector) this.mSpace.getAllGlyphs().clone();
        this.elementsToFade.remove(this.magWindow);
        final ClosedShape thisEndShape = n.getShape();
        final Glyph[] glyphs = n.getGlyphs();
        for (final Glyph glyph : glyphs) {
            this.elementsToFade.remove(glyph);
        }
        // compute layout of brought nodes
        final double thisEndBoundingCircleRadius = thisEndShape.getSize();
        // distance between two rings
        final double RING_STEP = 2 * thisEndBoundingCircleRadius;
        final LEdge[] arcs = n.getAllArcs();
        // sort them according to distance from start node
        // (so as to try to keep the closest ones closer to the start node)
        Arrays.sort(arcs, new DistanceComparator(n));
        final Hashtable node2broughtPosition = new Hashtable();
        final RingManager rm = new RingManager();
        for (final LEdge arc : arcs) {
            // ignore arcs that start and end at this node
            if (arc.isLoop()) {
                continue;
            }
            // process all others
            final LNode otherEnd = arc.getOtherEnd(n);
            final ClosedShape otherEndShape = otherEnd.getShape();
            final double d =
                Math.sqrt(Math.pow(otherEndShape.vx - thisEndShape.vx, 2)
                    + Math.pow(otherEndShape.vy - thisEndShape.vy, 2));
            // use this end's shape size when getting ring because we're rescaling
            // brought nodes to match the size of the current node
            final Ring ring =
                rm.getRing(
                    Math.atan2(otherEndShape.vy - thisEndShape.vy, otherEndShape.vx - thisEndShape.vx),
                    thisEndShape.getSize(),
                    RING_STEP);
            final double bd = ring.rank * RING_STEP;
            final double ratio = bd / d;
            final double bx = thisEndShape.vx + ratio * (otherEndShape.vx - thisEndShape.vx);
            final double by = thisEndShape.vy + ratio * (otherEndShape.vy - thisEndShape.vy);
            node2broughtPosition.put(otherEnd, new Point2D.Double(bx, by));
        }
        // actually bring the arcs and nodes (animation)
        for (final LEdge arc : arcs) {
            // ignore arcs that start and end at this node
            if (arc.isLoop()) {
                continue;
            }
            final LNode otherEnd = arc.getOtherEnd(n);
            final ClosedShape otherEndShape = otherEnd.getShape();
            bring(
                arc,
                otherEnd,
                thisEndShape.vx,
                thisEndShape.vy,
                otherEndShape.vx,
                otherEndShape.vy,
                thisEndShape.getSize(),
                node2broughtPosition);
        }
        // make other elements translucent (fade them out)
        this.alphaOfElementsToFade = new float[this.elementsToFade.size()];
        for (int i = 0; i < this.elementsToFade.size(); i++) {
            try {
                final Translucent t = (Translucent) this.elementsToFade.elementAt(i);
                this.alphaOfElementsToFade[i] = t.getTranslucencyValue();
                //XXX:TBW animate only if total number of objects is not too high
                final Animation a =
                    this.animator.getAnimationFactory().createTranslucencyAnim(
                        BRING_ANIM_DURATION,
                        t,
                        FADED_ELEMENTS_TRANSLUCENCY,
                        false,
                        IdentityInterpolator.getInstance(),
                        null);
                this.animator.startAnimation(a, true);
            } catch (final ClassCastException e) {
                e.printStackTrace();
            }
        }
    }

    void endBringAndGo(final Glyph g) {
        this.isBringingAndGoing = false;
        if (this.elementsToFade == null) {
            // there is nothing to end, actually
            return;
        }
        // new camera location (if g != null)
        Point2D.Double newCameraLocation = null;
        if (!this.broughtElements.isEmpty()) {
            for (int i = this.broughtElements.size() - 1; i >= 0; i--) {
                final Point2D.Double l = sendBack((BroughtElement) this.broughtElements.elementAt(i), g);
                if (l != null) {
                    newCameraLocation = l;
                }
            }
        }
        for (int i = 0; i < this.elementsToFade.size(); i++) {
            try {
                //XXX:TBW animate only if total number of objects is not too high
                final Animation a =
                    this.animator.getAnimationFactory().createTranslucencyAnim(
                        BRING_ANIM_DURATION,
                        (Translucent) this.elementsToFade.elementAt(i),
                        this.alphaOfElementsToFade[i],
                        false,
                        IdentityInterpolator.getInstance(),
                        null);
                this.animator.startAnimation(a, true);
            } catch (final ClassCastException e) {
                e.printStackTrace();
            }
        }
        this.elementsToFade.clear();
        if (newCameraLocation != null) {
            final Animation a =
                this.animator.getAnimationFactory().createCameraTranslation(
                    BRING_ANIM_DURATION,
                    this.mainCamera,
                    newCameraLocation,
                    false,
                    SlowInSlowOutInterpolator.getInstance(),
                    null);
            this.animator.startAnimation(a, true);
        }
    }

    void bring(
        final LEdge arc,
        final LNode node,
        final double sx,
        final double sy,
        final double ex,
        final double ey,
        final double size,
        final Hashtable node2broughtPosition)
    {
        this.broughtElements.add(BroughtElement.rememberPreviousState(node));
        this.broughtElements.add(BroughtElement.rememberPreviousState(arc));
        // deal with node glyphs
        final ClosedShape nodeShape = node.getShape();

        if (size < nodeShape.getSize()) {
            final Animation szAnim =
                this.animator.getAnimationFactory().createGlyphSizeAnim(
                    BRING_ANIM_DURATION,
                    nodeShape,
                    size,
                    false,
                    SlowInSlowOutInterpolator.getInstance(),
                    null);
            this.animator.startAnimation(szAnim, true);
            final VText label = node.getLabel();
            if (label != null) {
                label.setScale((float) (size / nodeShape.getSize()));
            }
        }

        this.elementsToFade.remove(nodeShape);
        final Point2D.Double bposition = (Point2D.Double) node2broughtPosition.get(node);
        final Point2D.Double translation = new Point2D.Double(bposition.x - nodeShape.vx, bposition.y - nodeShape.vy);
        Glyph[] glyphs = node.getGlyphs();
        for (final Glyph glyph : glyphs) {
            this.elementsToFade.remove(glyph);
            final Animation posAnim =
                this.animator.getAnimationFactory().createGlyphTranslation(
                    BRING_ANIM_DURATION,
                    glyph,
                    translation,
                    true,
                    SlowInSlowOutInterpolator.getInstance(),
                    null);
            this.animator.startAnimation(posAnim, true);
        }

        // deal with the edge's spline
        DPath spline = arc.getSpline();
        this.elementsToFade.remove(spline);
        Point2D.Double asp = spline.getStartPoint();
        Point2D.Double aep = spline.getEndPoint();
        Point2D.Double sp, ep;
        if (Math.sqrt(Math.pow(asp.x - ex, 2) + Math.pow(asp.y - ey, 2)) < Math.sqrt(Math.pow(asp.x - sx, 2)
            + Math.pow(asp.y - sy, 2)))
        {
            sp = new Point2D.Double(bposition.x, bposition.y);
            ep = new Point2D.Double(sx, sy);
        } else {
            sp = new Point2D.Double(sx, sy);
            ep = new Point2D.Double(bposition.x, bposition.y);
        }
        this.mSpace.above(spline, this.boundingBox);
        Point2D.Double[] flatCoords = DPath.getFlattenedCoordinates(spline, sp, ep, true);
        Animation a =
            this.animator.getAnimationFactory().createPathAnim(
                BRING_ANIM_DURATION,
                spline,
                flatCoords,
                false,
                SlowInSlowOutInterpolator.getInstance(),
                null);
        this.animator.startAnimation(a, true);
        // deal with the other glyphs that make the edge
        glyphs = arc.getGlyphs();
        for (final Glyph glyph : glyphs) {
            if (glyph != spline) {
                this.elementsToFade.remove(glyph);
                if (glyph instanceof VText) {
                    // put any label at the center of the edge (simplest thing we can do)
                    a =
                        this.animator.getAnimationFactory().createGlyphTranslation(
                            BRING_ANIM_DURATION,
                            glyph,
                            new Point2D.Double(bposition.x - sx, bposition.y - sy),
                            true,
                            SlowInSlowOutInterpolator.getInstance(),
                            null);
                    this.animator.startAnimation(a, true);
                } else {
                    // probably a tail or head decoration ; just hide it for now, we don't know how to transform them correctly
                    glyph.setVisible(false);
                }
            }
        }
        final LEdge[] otherArcs = node.getOtherArcs(arc);
        Glyph oe;
        for (final LEdge otherArc : otherArcs) {
            this.broughtElements.add(BroughtElement.rememberPreviousState(otherArc));
            spline = otherArc.getSpline();
            this.elementsToFade.remove(spline);
            asp = spline.getStartPoint();
            aep = spline.getEndPoint();
            if (node2broughtPosition.containsKey(otherArc.getTail())
                && node2broughtPosition.containsKey(otherArc.getHead()))
            {
                sp = (Point2D.Double) node2broughtPosition.get(otherArc.getTail());
                ep = (Point2D.Double) node2broughtPosition.get(otherArc.getHead());
            } else {
                oe = otherArc.getOtherEnd(node).getShape();
                if (Math.sqrt(Math.pow(asp.x - ex, 2) + Math.pow(asp.y - ey, 2)) <= Math.sqrt(Math.pow(aep.x - ex, 2)
                    + Math.pow(aep.y - ey, 2)))
                {
                    sp = new Point2D.Double(bposition.x, bposition.y);
                    ep = oe.getLocation();
                } else {
                    sp = oe.getLocation();
                    ep = new Point2D.Double(bposition.x, bposition.y);
                }
            }
            flatCoords = DPath.getFlattenedCoordinates(spline, sp, ep, true);
            this.mSpace.above(spline, this.boundingBox);
            a =
                this.animator.getAnimationFactory().createPathAnim(
                    BRING_ANIM_DURATION,
                    spline,
                    flatCoords,
                    false,
                    SlowInSlowOutInterpolator.getInstance(),
                    null);
            this.animator.startAnimation(a, true);
            spline.setTranslucencyValue(SECOND_STEP_TRANSLUCENCY);
        }
    }

    Point2D.Double sendBack(final BroughtElement be, final Glyph g) {
        this.broughtElements.remove(be);
        return be.restorePreviousState(BRING_ANIM_DURATION, g);
    }

    /* ----------------------- Link sliding navigation ----------------------------------- */

    static final int SLIDER_CURSOR_SIZE = 6;
    static final Color SLIDER_CURSOR_FILL = Color.WHITE;

    static final int SELECTION_RADIUS = 200;
    static final Color SELECTION_RADIUS_COLOR = Color.RED;

    boolean isLinkSliding = false;

    public boolean isLinkSliding() {
        return this.isLinkSliding;
    }

    LinkSliderCalc[] lscs;
    int lsci = -1;

    DPath slidingLink;
    Color slidingLinkActualColor = null;
    Point2D mPos = new Point2D.Double();

    SICircle slideCursor, selectionRadius;
    Point2D cPos;

    LNode closestNode;

    int screen_cursor_x, screen_cursor_y;
    Robot awtRobot;

    Point2D mtPos = new Point2D.Double();

    public void attemptLinkSliding(final double press_vx, final double press_vy, final int scr_x, final int scr_y) {
        final double vieww = this.mainView.getVisibleRegionWidth(this.mainCamera);
        this.lsci = 0;
        if (this.lstruct != null) {
            this.closestNode = this.lstruct.nodes[0];
            ClosedShape nodeShape = this.closestNode.getShape();
            double shortestDistance =
                Math.sqrt(Math.pow(nodeShape.vx - press_vx, 2) + Math.pow(nodeShape.vy - press_vy, 2));
            double distance;
            for (int i = 1; i < this.lstruct.nodes.length; i++) {
                nodeShape = this.lstruct.nodes[i].getShape();
                distance = Math.sqrt(Math.pow(nodeShape.vx - press_vx, 2) + Math.pow(nodeShape.vy - press_vy, 2));
                if (distance < shortestDistance) {
                    this.closestNode = this.lstruct.nodes[i];
                    shortestDistance = distance;
                }
            }
            if (shortestDistance < 2 * this.closestNode.getShape().getSize()) {
                // if clicked near a node, select edge connected to this node closest to the click point
                final LEdge[] arcs = this.closestNode.getAllArcs();
                if (arcs.length == 0) {
                    return;
                }
                this.lscs = new LinkSliderCalc[arcs.length];
                this.slidingLink = arcs[0].getSpline();
                this.lscs[0] = new LinkSliderCalc(this.slidingLink, vieww);
                this.mPos.setLocation(press_vx, press_vy);
                this.lscs[0].updateMousePosition(this.mPos);
                this.cPos = this.lscs[0].getPositionAlongPath();
                shortestDistance =
                    Math.sqrt(Math.pow(this.cPos.getX() - this.mPos.getX(), 2)
                        + Math.pow(this.cPos.getY() - this.mPos.getY(), 2));
                for (int i = 1; i < arcs.length; i++) {
                    this.lscs[i] = new LinkSliderCalc(arcs[i].getSpline(), vieww);
                    this.lscs[i].updateMousePosition(this.mPos);
                    this.cPos = this.lscs[i].getPositionAlongPath();
                    distance =
                        Math.sqrt(Math.pow(this.cPos.getX() - this.mPos.getX(), 2)
                            + Math.pow(this.cPos.getY() - this.mPos.getY(), 2));
                    if (distance < shortestDistance) {
                        shortestDistance = distance;
                        this.slidingLink = arcs[i].getSpline();
                        this.lsci = i;
                    }
                }
                startLinkSliding(press_vx, press_vy, scr_x, scr_y);
                return;
            }
        }
        // else select the edge hovered by the cursor (if any) - works even if no knowledge about logical structure
        this.closestNode = null;
        final Vector pum = this.mainView.getCursor().getPicker().getIntersectingPaths(this.mainCamera, 10);
        if (pum.size() > 0) {
            this.slidingLink = (DPath) pum.firstElement();
            this.lscs = new LinkSliderCalc[1];
            this.lscs[this.lsci] = new LinkSliderCalc(this.slidingLink, vieww);
            startLinkSliding(press_vx, press_vy, scr_x, scr_y);
        }
    }

    public void startLinkSliding(final double press_vx, final double press_vy, final int px, final int py) {
        this.mainView.getCursor().setVisibility(false);
        this.isLinkSliding = true;
        this.screen_cursor_x = px + this.panelWidth / 2;
        this.screen_cursor_y = py + this.panelHeight / 2;
        this.mainView.getPanel().setNoEventCoordinates(this.panelWidth / 2, this.panelHeight / 2);
        try {
            this.awtRobot = new Robot();
        } catch (final java.awt.AWTException e) {
            e.printStackTrace();
        }
        // chosen link
        this.slidingLinkActualColor = this.slidingLink.getColor();
        this.slidingLink.setColor(ConfigManager.HIGHLIGHT_COLOR);
        // add cursor on link
        this.slideCursor =
            new SICircle(press_vx, press_vy, 0, SLIDER_CURSOR_SIZE, SLIDER_CURSOR_FILL, ConfigManager.HIGHLIGHT_COLOR);
        this.slideCursor.setStroke(new BasicStroke(2f));
        //slideCursor.setFilled(false);
        this.mSpace.addGlyph(this.slideCursor);
        // display selection radius, circular zone that allows for arc switching when starting from a node
        if (this.closestNode != null) {
            this.selectionRadius =
                new SICircle(
                    this.closestNode.getShape().vx,
                    this.closestNode.getShape().vy,
                    0,
                    SELECTION_RADIUS,
                    Color.WHITE,
                    SELECTION_RADIUS_COLOR);
            this.selectionRadius.setFilled(false);
            this.selectionRadius.setStroke(new BasicStroke(2f));
            this.mSpace.addGlyph(this.selectionRadius);
        }
        // center camera on selection
        final Animation a =
            this.animator.getAnimationFactory().createCameraTranslation(
                200,
                this.mainCamera,
                new Point2D.Double(press_vx, press_vy),
                false,
                SlowInSlowOutInterpolator.getInstance(),
                new EndAction() {
                    @Override
                    public void execute(final Object subject, final Animation.Dimension dimension) {
                        linkSlider(press_vx, press_vy, true);
                    }
                });
        this.animator.startAnimation(a, false);
    }

    public void linkSlider(final double vx, final double vy, final boolean centerCursor) {
        if (this.lscs == null) {
            // endLinkSliding might have been called before the call to
            // linkSlider(...) gets fired by the animation at the end of startLinkSliding,
            // e.g., when clicking in LinkSliding mode
            return;
        }
        final boolean withinSelectionRadius = this.mainView.getCursor().getPicker().isPicked(this.selectionRadius);
        this.mPos.setLocation(vx, vy);
        this.lscs[this.lsci].updateMousePosition(this.mPos);
        if (!withinSelectionRadius || centerCursor) {
            // constrained sliding on link
            this.awtRobot.mouseMove(this.screen_cursor_x, this.screen_cursor_y);
        } else {
            // relatively free movements in link selection area around nodes
            this.mtPos.setLocation(vx, vy);
            int newlsci = this.lsci;
            Point2D tPos = this.lscs[this.lsci].getPositionAlongPath();
            double shortestDistance =
                Math.sqrt(Math.pow(tPos.getX() - this.mtPos.getX(), 2)
                    + Math.pow(this.cPos.getY() - this.mtPos.getY(), 2));
            double distance;
            for (int i = 0; i < this.lsci; i++) {
                this.lscs[i].updateMousePosition(this.mtPos);
                tPos = this.lscs[i].getPositionAlongPath();
                distance =
                    Math.sqrt(Math.pow(tPos.getX() - this.mtPos.getX(), 2)
                        + Math.pow(this.cPos.getY() - this.mtPos.getY(), 2));
                if (distance < shortestDistance) {
                    shortestDistance = distance;
                    newlsci = i;
                }
            }
            for (int i = this.lsci + 1; i < this.lscs.length; i++) {
                this.lscs[i].updateMousePosition(this.mtPos);
                tPos = this.lscs[i].getPositionAlongPath();
                distance =
                    Math.sqrt(Math.pow(tPos.getX() - this.mtPos.getX(), 2)
                        + Math.pow(this.cPos.getY() - this.mtPos.getY(), 2));
                if (distance < shortestDistance) {
                    shortestDistance = distance;
                    newlsci = i;
                }
            }
            if (newlsci != this.lsci) {
                this.slidingLink.setColor(this.slidingLinkActualColor);
                this.lsci = newlsci;
                this.slidingLink = this.lscs[this.lsci].getPath();
                this.slidingLinkActualColor = this.slidingLink.getColor();
                this.slidingLink.setColor(ConfigManager.HIGHLIGHT_COLOR);
            }
        }
        this.cPos = this.lscs[this.lsci].getPositionAlongPath();
        this.slideCursor.moveTo(Math.round(this.cPos.getX()), Math.round(this.cPos.getY()));
        this.mainCamera.moveTo(Math.round(this.cPos.getX()), Math.round(this.cPos.getY()));
        //mainCamera.setAltitude((float)(Camera.DEFAULT_FOCAL/lsc.getScale() - Camera.DEFAULT_FOCAL));
    }

    public void endLinkSliding() {
        this.mainView.getPanel().setNoEventCoordinates(ViewPanel.NO_COORDS, ViewPanel.NO_COORDS);
        this.mainView.getCursor().setVisibility(true);
        this.mSpace.removeGlyph(this.slideCursor);
        if (this.selectionRadius != null) {
            this.mSpace.removeGlyph(this.selectionRadius);
            this.selectionRadius = null;
        }
        this.slidingLink.setColor(this.slidingLinkActualColor);
        this.slidingLink = null;
        this.closestNode = null;
        this.isLinkSliding = false;
        this.lscs = null;
        this.lsci = -1;
        this.awtRobot = null;
    }

}

class ZGRGlassPane extends JComponent {

    static final int BAR_WIDTH = 200;
    static final int BAR_HEIGHT = 10;

    static final AlphaComposite GLASS_ALPHA = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.65f);
    static final Color MSG_COLOR = Color.DARK_GRAY;
    GradientPaint PROGRESS_GRADIENT = new GradientPaint(0, 0, Color.ORANGE, 0, BAR_HEIGHT, Color.BLUE);

    String msg = Messages.EMPTY_STRING;
    int msgX = 0;
    int msgY = 0;

    int completion = 0;
    int prX = 0;
    int prY = 0;
    int prW = 0;

    GraphicsManager grMngr;

    ZGRGlassPane(final GraphicsManager grMngr) {
        super();
        this.grMngr = grMngr;
        addMouseListener(new MouseAdapter() {
        });
        addMouseMotionListener(new MouseMotionAdapter() {
        });
        addKeyListener(new KeyAdapter() {
        });
    }

    void setProgress(final int c) {
        this.completion = c;
        this.prX = this.grMngr.panelWidth / 2 - BAR_WIDTH / 2;
        this.prY = this.grMngr.panelHeight / 2 - BAR_HEIGHT / 2;
        this.prW = (int) (BAR_WIDTH * ((float) this.completion) / 100.0f);
        this.PROGRESS_GRADIENT =
            new GradientPaint(0, this.prY, Color.LIGHT_GRAY, 0, this.prY + BAR_HEIGHT, Color.DARK_GRAY);
        repaint(this.prX, this.prY, BAR_WIDTH, BAR_HEIGHT);
    }

    void setMessage(final String m) {
        this.msg = m;
        this.msgX = this.grMngr.panelWidth / 2 - BAR_WIDTH / 2;
        this.msgY = this.grMngr.panelHeight / 2 - BAR_HEIGHT / 2 - 10;
        repaint(this.msgX, this.msgY - 50, 200, 70);
    }

    @Override
    protected void paintComponent(final Graphics g) {
        final Graphics2D g2 = (Graphics2D) g;
        final Rectangle clip = g.getClipBounds();
        g2.setComposite(GLASS_ALPHA);
        g2.setColor(Color.WHITE);
        g2.fillRect(clip.x, clip.y, clip.width, clip.height);
        g2.setComposite(AlphaComposite.Src);
        if (this.msg != Messages.EMPTY_STRING) {
            g2.setColor(MSG_COLOR);
            g2.setFont(ConfigManager.defaultFont);
            g2.drawString(this.msg, this.msgX, this.msgY);
        }
        g2.setPaint(this.PROGRESS_GRADIENT);
        g2.fillRect(this.prX, this.prY, this.prW, BAR_HEIGHT);
        g2.setColor(MSG_COLOR);
        g2.drawRect(this.prX, this.prY, BAR_WIDTH, BAR_HEIGHT);
    }

}

class ZP2LensAction implements EndAction {

    GraphicsManager grMngr;

    public ZP2LensAction(final GraphicsManager gm) {
        this.grMngr = gm;
    }

    @Override
    public void execute(final Object subject, final Animation.Dimension dimension) {
        ((Lens) subject).getOwningView().setLens(null);
        ((Lens) subject).dispose();
        this.grMngr.setMagFactor(GraphicsManager.DEFAULT_MAG_FACTOR);
        this.grMngr.lens.dispose();
        this.grMngr.lens = null;
        this.grMngr.setLens(GraphicsManager.NO_LENS);
    }

}

class RingManager {

    Ring[] rings = new Ring[0];

    Ring getRing(double direction, final double size, final double ringStep) {
        // normalize direction in [0,2Pi[
        if (direction < 0) {
            direction = 2 * Math.PI + direction;
        }
        // look for a ring where the new object could be placed, starting with the innermost one
        for (int i = 0; i < this.rings.length; i++) {
            final double a = Math.abs(Math.atan2(size, this.rings[i].rank * ringStep));
            if (!this.rings[i].intersectsConeOfInfluence(direction - a, direction + a)) {
                this.rings[i].addNode(direction - a, direction + a);
                return this.rings[i];
            }
        }
        // if couldn't find any room, create a new ring
        final Ring r = createNewRing();
        final double a = Math.abs(Math.atan2(size, ringStep));
        r.addNode(direction - a, direction + a);
        return r;
    }

    private Ring createNewRing() {
        final Ring[] tr = new Ring[this.rings.length + 1];
        System.arraycopy(this.rings, 0, tr, 0, this.rings.length);
        tr[this.rings.length] = new Ring(tr.length);
        this.rings = tr;
        return this.rings[this.rings.length - 1];
    }

}

class Ring {

    /* rank of this ring (starts at 1) */
    int rank;
    //  /* nodes on this ring */
    //  LNode[] nodes = new LNode[0];
    //  /* nodes on this ring */
    //  Point2D.Double[] broughtPositions = new Point2D.Double[0];
    //  /* cones of influence, for each item, first element is the smallest angle in [0, 2Pi[, second the largest angle in [0, 2Pi[ */
    double[][] cones = new double[0][2];

    Ring(final int r) {
        this.rank = r;
    }

    //	void addNode(/*LNode n, Point2D.Double p,*/ double a1, double a2){
    void addNode(double a1, double a2) {
        //		// add node
        //		LNode[] ta = new LNode[nodes.length+1];
        //		System.arraycopy(nodes, 0, ta, 0, nodes.length);
        //		ta[nodes.length] = n;
        //		nodes = ta;
        //		// add node
        //		Point2D.Double[] tp = new Point2D.Double[broughtPositions.length+1];
        //		System.arraycopy(broughtPositions, 0, tp, 0, broughtPositions.length);
        //		tp[nodes.length] = p;
        //		broughtPositions = tp;
        // compute its cone of influence
        final double[][] tc = new double[this.cones.length + 1][2];
        System.arraycopy(this.cones, 0, tc, 0, this.cones.length);
        // normalize angles in [0,2Pi[
        if (a1 < 0) {
            a1 = 2 * Math.PI + a1;
        }
        if (a2 < 0) {
            a2 = 2 * Math.PI + a2;
        }
        tc[this.cones.length][0] = Math.min(a1, a2);
        tc[this.cones.length][1] = Math.max(a1, a2);
        this.cones = tc;
    }

    boolean intersectsConeOfInfluence(final double a1, final double a2) {
        for (final double[] cone : this.cones) {
            if (a2 > cone[0] && a1 < cone[1]) {
                return true;
            }
        }
        return false;
    }

}

class DistanceComparator implements java.util.Comparator {

    LNode centerNode;
    Glyph centerShape;

    DistanceComparator(final LNode cn) {
        this.centerNode = cn;
        this.centerShape = cn.getShape();
    }

    @Override
    public int compare(final Object o1, final Object o2) {
        final Glyph n1 = ((LEdge) o1).getOtherEnd(this.centerNode).getShape();
        final Glyph n2 = ((LEdge) o2).getOtherEnd(this.centerNode).getShape();
        final double d1 = Math.pow(this.centerShape.vx - n1.vx, 2) + Math.pow(this.centerShape.vy - n1.vy, 2);
        final double d2 = Math.pow(this.centerShape.vx - n2.vx, 2) + Math.pow(this.centerShape.vy - n2.vy, 2);
        if (d1 < d2) {
            return -1;
        } else if (d1 > d2) {
            return 1;
        } else {
            return 0;
        }
    }

}
