/*   FILE: PeriodicActionManager.java
 *   DATE OF CREATION:   Thu Jan 09 14:14:35 2003
 *   Copyright (c) Emmanuel Pietriga, 2002. All Rights Reserved
 *   Copyright (c) INRIA, 2004-2011. All Rights Reserved
 *   Licensed under the GNU LGPL. For full terms see the file COPYING.
 *
 * $Id$
 */

package net.claribole.zgrviewer;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;

import fr.inria.zvtm.engine.*;
import fr.inria.zvtm.glyphs.*;

class PeriodicActionManager implements Runnable, MouseMotionListener, Java2DPainter {

    static int SLEEP_TIME = 500; // check for tooltip changes every 1.0s
    static int TOOLTIP_TIME = 1000; // tooltip info should appear only if mouse is idle for at least 1.5s

    static Color TP_BACKGROUND = new Color(255, 255, 147);
    static Color TP_FOREGROUND = Color.black;
    static Font TP_FONT = new Font("Dialog", Font.PLAIN, 10);
    static int TP_PADDING = 4;
    static int TP_MARGIN = 15;
    private boolean invalidBounds = true;

    GraphicsManager grMngr;

    Thread runTP;

    long lastMouseMoved = System.currentTimeMillis();

    Glyph tippedGlyph;
    String tipLabel;
    int lX, lY, rX, rY, rW, rH;

    boolean updatePalette = false;

    PeriodicActionManager(final GraphicsManager gm) {
        this.grMngr = gm;
    }

    public void start() {
        this.runTP = new Thread(this);
        this.runTP.setPriority(Thread.MIN_PRIORITY);
        this.runTP.start();
    }

    public synchronized void stop() {
        this.runTP = null;
        notify();
    }

    @Override
    public void run() {
        final Thread me = Thread.currentThread();
        while (this.runTP == me) {
            updateTooltip();
            checkToolPalette();
            try {
                Thread.sleep(SLEEP_TIME); //sleep ... ms  
            } catch (final InterruptedException e) {
                return;
            }
        }
    }

    void updateTooltip() {
        if ((System.currentTimeMillis() - this.lastMouseMoved) > TOOLTIP_TIME) {
            final Glyph g = this.grMngr.mainView.getPanel().lastGlyphEntered();
            if (g != null && g != this.grMngr.boundingBox && this.tippedGlyph != g) {
                this.tippedGlyph = g;
                if (this.tippedGlyph.getOwner() != null && this.tippedGlyph.getOwner() instanceof LElem) {
                    this.tipLabel = ((LElem) this.tippedGlyph.getOwner()).getTooltip(this.tippedGlyph);
                }
                if (this.tipLabel != null && this.tipLabel.length() > 0) {
                    this.lX = this.grMngr.mainView.mouse.getPanelXCoordinate() + TP_MARGIN;
                    this.lY = this.grMngr.mainView.mouse.getPanelYCoordinate() + TP_MARGIN;
                    this.invalidBounds = true;
                    this.grMngr.vsm.repaint();
                }
            }
        }
    }

    void removeTooltip() {
        this.tipLabel = null;
        this.tippedGlyph = null;
        this.invalidBounds = true;
        this.grMngr.vsm.repaint();
    }

    void computeTipRectangle(final int labelWidth, final int labelHeight) {
        this.rX = this.lX - TP_PADDING;
        this.rY = this.lY - labelHeight;
        this.rW = labelWidth + TP_PADDING + TP_PADDING;
        this.rH = labelHeight + TP_PADDING;
    }

    @Override
    public void paint(final Graphics2D g2d, final int viewWidth, final int viewHeight) {
        if (this.tipLabel != null) {
            final Font origFont = g2d.getFont();
            g2d.setFont(TP_FONT);
            if (this.invalidBounds) {
                final Rectangle2D r2d = g2d.getFontMetrics().getStringBounds(this.tipLabel, g2d);
                computeTipRectangle((int) r2d.getWidth(), (int) r2d.getHeight());
                this.invalidBounds = false;
            }
            g2d.setColor(TP_BACKGROUND);
            g2d.fillRect(this.rX, this.rY, this.rW, this.rH);
            g2d.setColor(TP_FOREGROUND);
            g2d.drawRect(this.rX, this.rY, this.rW, this.rH);
            g2d.drawString(this.tipLabel, this.lX, this.lY);
            g2d.setFont(origFont);
        }
    }

    @Override
    public void mouseDragged(final MouseEvent e) {
        this.lastMouseMoved = System.currentTimeMillis();
        removeTooltip();
    }

    @Override
    public void mouseMoved(final MouseEvent e) {
        this.lastMouseMoved = System.currentTimeMillis();
        removeTooltip();
    }

    void requestToolPaletteRelocation() {
        this.updatePalette = true;
    }

    void checkToolPalette() {
        if (this.updatePalette) {
            this.grMngr.tp.updateHiddenPosition();
            this.updatePalette = false;
        }
    }

}
