/*   FILE: ZGRViewer.java
 *   DATE OF CREATION:   Wed Aug 30 12:02:31 2006
 *   Copyright (c) INRIA, 2006-2011. All Rights Reserved
 *   Licensed under the GNU LGPL. For full terms see the file COPYING.
 *
 *   $Id$
 */

package net.claribole.zgrviewer;

import java.awt.*;
import java.util.*;

import javax.swing.*;

import fr.inria.zvtm.engine.*;
import fr.inria.zvtm.glyphs.*;

public class ToolPalette {

    GraphicsManager grMngr;

    static final String PALETTE_SPACE_NAME = "tps";
    VirtualSpace paletteSpace;
    Camera paletteCamera;

    public static final String PLUGIN_MODE_PREFIX = "plugin";

    public static final String STD_NAV_MODE = "st";
    public static final String FL_NAV_MODE = "fl";
    public static final String DM_NAV_MODE = "dm";
    public static final String PL_NAV_MODE = "pl";
    public static final String HIGHLIGHT_MODE = "hi";
    public static final String BRING_AND_GO_MODE = "bg";
    public static final String LINK_SLIDING_MODE = "ls";
    public static final String EDIT_MODE = "ed";

    static final String[] BUILTIN_MODES = {
        STD_NAV_MODE,
        FL_NAV_MODE,
        DM_NAV_MODE,
        PL_NAV_MODE,
        HIGHLIGHT_MODE,
        BRING_AND_GO_MODE,
        LINK_SLIDING_MODE,
        EDIT_MODE };

    static final HashMap<String, String> BUILTIN_MODE_ICON_PATHS = new HashMap(8);
    static {
        BUILTIN_MODE_ICON_PATHS.put(STD_NAV_MODE, "/images/stdnav24b.png");
        BUILTIN_MODE_ICON_PATHS.put(FL_NAV_MODE, "/images/flnav24b.png");
        BUILTIN_MODE_ICON_PATHS.put(DM_NAV_MODE, "/images/dmnav24b.png");
        BUILTIN_MODE_ICON_PATHS.put(PL_NAV_MODE, "/images/plnav24b.png");
        BUILTIN_MODE_ICON_PATHS.put(HIGHLIGHT_MODE, "/images/hl24b.png");
        BUILTIN_MODE_ICON_PATHS.put(BRING_AND_GO_MODE, "/images/fl24b.png");
        BUILTIN_MODE_ICON_PATHS.put(LINK_SLIDING_MODE, "/images/ls24b.png");
        BUILTIN_MODE_ICON_PATHS.put(EDIT_MODE, "/images/edit24b.png");
    };

    VImage[] buttons;
    VImage[] selectedButtons;
    static final int VERTICAL_STEP_BETWEEN_ICONS = 30;

    VImage currentModeIcon = null;
    String currentMode = null;

    boolean visible = false;
    boolean paintPalette = true; // set to false temporarily during panel resizing operations ; used as an optimization indicator
    boolean enabled = true;

    public static final int ANIM_TIME = 200;
    public static final int TRIGGER_ZONE_WIDTH = 48;
    public static int TRIGGER_ZONE_HEIGHT = BUILTIN_MODES.length * (VERTICAL_STEP_BETWEEN_ICONS) + 24;

    short firstPluginModeIndex = -1;

    ToolPalette(final GraphicsManager gm) {
        this.grMngr = gm;
        loadBuiltinModes();
        loadPluginModes();
        // update original trigger zone height
        TRIGGER_ZONE_HEIGHT = this.buttons.length * (VERTICAL_STEP_BETWEEN_ICONS) + 24;
        selectDefaultMode();
    }

    void loadBuiltinModes() {
        this.paletteSpace = this.grMngr.vsm.addVirtualSpace(PALETTE_SPACE_NAME);
        this.paletteCamera = this.paletteSpace.addCamera();
        this.paletteCamera.setAltitude(0);
        // building list of built-in modes that plugins want to disable
        final Plugin[] plugins = this.grMngr.cfgMngr.plugins;
        final HashMap<String, Object> modesToDisable = new HashMap(2);
        if (plugins != null) {
            for (short i = 0; i < plugins.length; i++) {
                final String[] mtd = plugins[i].getDisabledModes();
                if (mtd != null && mtd.length > 0) {
                    for (final String bimode : mtd) {
                        modesToDisable.put(bimode, null);
                    }
                }
            }
        }
        // instantiating remaining built-in modes
        this.buttons = new VImage[BUILTIN_MODES.length - modesToDisable.size()];
        short i = 0;
        short j = 0;
        while (i < BUILTIN_MODES.length) {
            if (modesToDisable.containsKey(BUILTIN_MODES[i])) {
            } else {
                this.buttons[j] =
                    new VImage(0, -j * VERTICAL_STEP_BETWEEN_ICONS, 0, (new ImageIcon(this.getClass().getResource(
                        BUILTIN_MODE_ICON_PATHS.get(BUILTIN_MODES[i])))).getImage());
                //this.paletteSpace.addGlyph(this.buttons[j]);
                this.buttons[j].setOwner(BUILTIN_MODES[i]);
                j++;
            }
            i++;
        }
        this.firstPluginModeIndex = j;
    }

    void selectDefaultMode() {
        if (defaultPluginModeClassName != null && this.pluginsWithMode != null && this.pluginsWithMode.size() > 0) {
            for (final Short index : this.pluginsWithMode.keySet()) {
                final Plugin p = this.pluginsWithMode.get(index);
                if (p.getClass().getName().equals(defaultPluginModeClassName)) {
                    selectMode(this.buttons[index.shortValue()]);
                    return;
                }
            }
        }
        selectMode(this.buttons[0]);
    }

    public void setEnabled(final boolean b) {
        this.enabled = b;
        this.paletteCamera.setEnabled(b);
    }

    public boolean isEnabled() {
        return this.enabled;
    }

    public boolean isStdNavMode() {
        return this.currentMode.equals(STD_NAV_MODE);
    }

    public boolean isFadingLensNavMode() {
        return this.currentMode.equals(FL_NAV_MODE);
    }

    public boolean isDragMagNavMode() {
        return this.currentMode.equals(DM_NAV_MODE);
    }

    public boolean isProbingLensNavMode() {
        return this.currentMode.equals(PL_NAV_MODE);
    }

    public boolean isHighlightMode() {
        return this.currentMode.equals(HIGHLIGHT_MODE);
    }

    public boolean isBringAndGoMode() {
        return this.currentMode.equals(BRING_AND_GO_MODE);
    }

    public boolean isLinkSlidingMode() {
        return this.currentMode.equals(LINK_SLIDING_MODE);
    }

    public boolean isEditMode() {
        return this.currentMode.equals(EDIT_MODE);
    }

    public String getMode() {
        return this.currentMode;
    }

    static final BasicStroke SB_STROKE = new BasicStroke(2f);
    static final BasicStroke DB_STROKE = new BasicStroke(1f);

    public void selectMode(final VImage icon) {
        boolean newModeSelected = false;
        final String previousMode = this.currentMode;
        if (icon != this.currentModeIcon) {
            if (this.currentModeIcon != null) {
                // remove thick border from previously selected icon
                this.currentModeIcon.setStroke(DB_STROKE);
            }
            this.currentModeIcon = icon;
            // add thick border to newly selected icon
            this.currentModeIcon.setStroke(SB_STROKE);
            this.currentMode = (String) icon.getOwner();
            newModeSelected = true;
        }
        if (newModeSelected && previousMode != null) {
            // if a new button has been selected,
            // discard resources associated with old mode
            if (previousMode.equals(DM_NAV_MODE)) {
                this.grMngr.killDM();
            } else if (previousMode.equals(BRING_AND_GO_MODE)) {
                this.grMngr.exitBringAndGoMode();
            } else if (previousMode.equals(STD_NAV_MODE)) {
                this.grMngr.activateDynaSpot(false, false);
            } else if (previousMode.equals(EDIT_MODE)) {
                this.grMngr.geom.clearSplineEditingGlyphs();
            }
            // init new mode
            if (this.currentMode.equals(BRING_AND_GO_MODE)) {
                this.grMngr.enterBringAndGoMode();
            } else if (this.currentMode.equals(STD_NAV_MODE)) {
                if (ConfigManager.DYNASPOT) {
                    try {
                        this.grMngr.activateDynaSpot(true, false);
                    } catch (final NullPointerException ex) {
                    }
                }
            }
            // callbacks
            if (previousMode.startsWith(PLUGIN_MODE_PREFIX)) {
                // a plugin mode, notify the plugin
                getPlugin(Short.parseShort(previousMode.substring(PLUGIN_MODE_PREFIX.length()))).exitMode();
            }
            if (this.currentMode.startsWith(PLUGIN_MODE_PREFIX)) {
                // a plugin mode, notify the plugin
                getPlugin(Short.parseShort(this.currentMode.substring(PLUGIN_MODE_PREFIX.length()))).enterMode();
            }
        }
    }

    /* Called with false when resizing the main view to temporarily hide the palette
       until it actually gets relocated to the top-left corner of that window.
       It is then called with true.*/
    public void displayPalette(final boolean b) {
        if (this.paintPalette == b) {
            return;
        }
        for (final VImage button : this.buttons) {
            //button.setVisible(b);
        }
        this.paintPalette = b;
    }

    public void updateHiddenPosition() {
        /*
        final double[] wnes = this.grMngr.mainView.getVisibleRegion(this.paletteCamera);
        for (int i = 0; i < this.buttons.length; i++) {
            this.buttons[i].moveTo(wnes[0] - this.buttons[i].getWidth() / 2 + 1, wnes[1]
                - (i + 1)
                * VERTICAL_STEP_BETWEEN_ICONS);
        }
        */
        displayPalette(true);
    }

    public void show() {
        /*
        if (!this.visible) {
            this.visible = true;
            final Animation a =
                this.grMngr.vsm
                    .getAnimationManager()
                    .getAnimationFactory()
                    .createCameraTranslation(
                        ANIM_TIME,
                        this.paletteCamera,
                        new Point2D.Double(-this.buttons[0].getWidth() - 2, 0),
                        true,
                        SlowInSlowOutInterpolator.getInstance(),
                        null);
            this.grMngr.vsm.getAnimationManager().startAnimation(a, false);
            this.grMngr.mainView.setCursorIcon(Cursor.HAND_CURSOR);
            this.grMngr.mainView.setActiveLayer(2);
        }
        */
    }

    public void hide() {
        /*
        if (this.visible) {
            this.visible = false;
            final Animation a =
                this.grMngr.vsm
                    .getAnimationManager()
                    .getAnimationFactory()
                    .createCameraTranslation(
                        ANIM_TIME,
                        this.paletteCamera,
                        new Point2D.Double(this.buttons[0].getWidth() + 2, 0),
                        true,
                        SlowInSlowOutInterpolator.getInstance(),
                        null);
            this.grMngr.vsm.getAnimationManager().startAnimation(a, false);
            this.grMngr.mainView.setCursorIcon(Cursor.CUSTOM_CURSOR);
            this.grMngr.mainView.setActiveLayer(0);
        }
        */
    }

    /**
     * return false if palette is temporarily disabled
     */
    public boolean insidePaletteTriggerZone(final int jpx, final int jpy) {
        return (this.paintPalette && jpx < TRIGGER_ZONE_WIDTH && jpy < TRIGGER_ZONE_HEIGHT);
    }

    public boolean isShowing() {
        return this.visible;
    }

    public Camera getPaletteCamera() {
        return this.paletteCamera;
    }

    //public void showLogicalTools(){
    //    for (int i=4;i<=6;i++){
    //        if (!buttons[i].isSensitive()){buttons[i].setSensitivity(true);}
    //        if (!buttons[i].isVisible()){buttons[i].setVisible(true);}
    //    }
    //}
    //
    //public void hideLogicalTools(){
    //	if (isHighlightMode() || isBringAndGoMode() || isLinkSlidingMode()){
    //		// if a tool that makes needs to know about the logical structure is selected,
    //		// select something else as they are about to be disabled
    //		selectButton(buttons[0]);
    //	}
    //	for (int i=4;i<=6;i++){
    //		if (buttons[i].isSensitive()){buttons[i].setSensitivity(false);}
    //		if (buttons[i].isVisible()){buttons[i].setVisible(false);}
    //	}
    //}

    /* ---------------- plugin modes ------------------*/

    static String defaultPluginModeClassName = null;

    static void setDefaultPluginMode(final String pluginClassName) {
        defaultPluginModeClassName = pluginClassName;
    }

    HashMap<Short, Plugin> pluginsWithMode;

    void loadPluginModes() {
        final Plugin[] plugins = this.grMngr.cfgMngr.plugins;
        if (plugins == null) {
            return;
        }
        final Vector<Plugin> pwm = new Vector<Plugin>(plugins.length);
        for (final Plugin plugin : plugins) {
            if (plugin.hasMode()) {
                pwm.add(plugin);
            }
        }
        if (pwm.isEmpty()) {
            return;
        }
        this.pluginsWithMode = new HashMap<Short, Plugin>(pwm.size());
        final VImage[] nbuttons = new VImage[this.buttons.length + pwm.size()];
        System.arraycopy(this.buttons, 0, nbuttons, 0, this.buttons.length);
        this.buttons = nbuttons;
        for (int i = this.firstPluginModeIndex; i < this.firstPluginModeIndex + pwm.size(); i++) {
            final Plugin p = pwm.elementAt(i - this.firstPluginModeIndex);
            this.pluginsWithMode.put(new Short((short) i), p);
            this.buttons[i] = new VImage(0, -i * VERTICAL_STEP_BETWEEN_ICONS, 0, p.getModeIcon());
            this.buttons[i].setOwner(PLUGIN_MODE_PREFIX + String.valueOf(i));
            this.paletteSpace.addGlyph(this.buttons[i]);
        }
    }

    Plugin getPlugin(final short index) {
        return this.pluginsWithMode.get(new Short(index));
    }

}
